import numpy as np
import shutil
from glob import glob
import os, sys

pltg = os.environ["PLTG"]
pltg_data = os.environ["PLTG_DATA"]

import inspect, sys

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
target_dir = os.path.dirname(os.path.dirname(current_dir))
sys.path.insert(0, target_dir)

from datetime import datetime, timedelta
import time
from aggregate_glm_grids import aggregate_glm_grids
from lightningcast import utils
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
import traceback
import argparse
from subprocess import Popen, PIPE, call
import pickle

# ***make sure ps_pythonpath is using ML environment***
ps_pythonpath = os.environ["PS_PYTHONPATH"]

eternity = "30991231-000000"
############################################################################################################################

# /satbuf1_data/goes/grb/goes16/2018/2018_06_11_162/glm/L2/LCFA/*s2018162130[0-4]*


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-g",
        "--glmdir",
        help="data directory; default = /satbuf1_data/goes/grb/goes16/",
        type=str,
        default=["/satbuf1_data/goes/grb/goes16/"],
        nargs=1,
    )
    parser.add_argument(
        "-st",
        "--starttime",
        help="starttime (e.g., 20180810-160000; default = now [realtime mode])",
        type=str,
        default=[datetime.utcnow().strftime("%Y%m%d-%H%M%S")],
        nargs=1,
    )
    parser.add_argument(
        "-et",
        "--endtime",
        help="endtime (e.g., 20180811-000000; default = run forever [realtime mode])",
        type=str,
        default=[eternity],
        nargs=1,
    )
    parser.add_argument(
        "-o",
        "--outdir",
        help="default = " + pltg_data + "/lightning/realtime/GLM/goes_east/",
        type=str,
        default=[pltg_data + "/lightning/realtime/GLM/goes_east/"],
        nargs=1,
    )
    parser.add_argument(
        "-sl",
        "--slot",
        help="orbital slot. default='goes_east'",
        default=["goes_east"],
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-sr", "--spatial_res", help="default = 2.0", type=str, default=["2.0"], nargs=1
    )
    parser.add_argument(
        "-geo",
        "--geotiffs",
        help="create geotiffs of gridded_products",
        action="store_true",
    )
    parser.add_argument(
        "-ru",
        "--re_upload",
        help="Full path to re_upload executable. Will execute after GTiff is written.",
        default=[None],
        type=str,
        nargs=1,
    )

    args = parser.parse_args()

    drive_make_GLM_grids(
        glmdir=args.glmdir[0],
        starttime=args.starttime[0],
        endtime=args.endtime[0],
        outdir=args.outdir[0],
        slot=args.slot[0],
        spatial_res=args.spatial_res[0],
        geotiffs=args.geotiffs,
        re_upload=args.re_upload[0],
    )


################################################################################################################################################
def drive_make_GLM_grids(
    glmdir="/satbuf1_data/goes/grb/goes16/",
    starttime=datetime.utcnow().strftime("%Y%m%d-%H%M%S"),
    endtime=eternity,
    outdir=f"{pltg_data}/lightning/realtime/GLM/goes_east/",
    spatial_res="2.0",
    slot="goes_east",
    geotiffs=False,
    re_upload=None,
):

    logDict = logging_def(logfile=None)
    dictConfig(logDict)
    # send stderr to logger
    stderr_logger = logging.getLogger("STDERR")
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl

    logging.info("Process started")

    if slot != "goes_east" and slot != "goes_west":
        logging.error(f"slot must be goes_east or goes_west, not {slot}")
        sys.exit()

    # some constants and paths
    dx = dy = spatial_res  # in km
    if dx != "2.0" and remap_res != "10.0":
        logging.error("spatial_res must be 2.0 or 10.0")
        sys.exit()

    if glmdir[-1] != "/":
        glmdir += "/"

    if endtime == eternity:
        # we're in realtime mode
        assert "realtime" in outdir
        outroot = outdir  # f"{pltg_data}/lightning/realtime/GLM/{slot}/"
        realtime = True
        raw_outdir = f"{outroot}/raw/"
    else:
        realtime = False
        # we're doing archive cases
        outroot = outdir  # this should be something like /data/common/GLM_training/ Then we append YYYYMMDD/ or raw/
        raw_outdir = f"{outdir}/raw/"

    try:
        utils.mkdir_p(outroot)
    except (OSError, IOError) as err:
        logging.error(traceback.format_exc())
        sys.exit()

    epoch = datetime.utcfromtimestamp(0)  # 1970/01/01 00:00

    try:
        dt = datetime.strptime(starttime, "%Y%m%d-%H%M%S")
        edt = datetime.strptime(endtime, "%Y%m%d-%H%M%S")
    except ValueError as err:
        logging.error(traceback.format_exc())
        return False

    lastfile = ""

    while dt < edt:

        if realtime:
            dt = datetime.utcnow()
            oneMinAgo = dt - timedelta(minutes=1)
        else:
            oneMinAgo = dt

        latest_file = glob(
            glmdir + oneMinAgo.strftime("%Y/%Y_%m_%d_%j/glm/L2/LCFA/*s%Y%j%H%M40*")
        )

        if len(latest_file):
            logging.info(f"Latest file is {latest_file[0]}")
            remotedir = os.path.dirname(latest_file[0]) + "/"
            latest_file = os.path.basename(latest_file[0])
            if latest_file != lastfile:  # i.e., not already processed
                logging.info(f"Latest file is {latest_file}")

                files_to_process = remotedir + oneMinAgo.strftime("*s%Y%j%H%M*")

                # ps_pythonpath must be using ML environment!
                if "ussamoa" in outroot:
                    logging.info(
                        f"{ps_pythonpath}/bin/python {pltg}/lightningcast/glm/glmtools/examples/grid/make_GLM_grids.py \
                        -o {raw_outdir}/{{start_time:%Y/%b/%d}}/{{dataset_name}} --fixed_grid --split_events \
                        --goes_position {slot[-4:]} --dx={dx} --dy={dy} --ctr_lat -13 --ctr_lon -167.5 --width 1250 --height 900 {files_to_process}"
                    )
                else:
                    logging.info(
                        f"{ps_pythonpath}/bin/python {pltg}/lightningcast/glm/glmtools/examples/grid/make_GLM_grids.py \
                         -o {raw_outdir}/{{start_time:%Y/%b/%d}}/{{dataset_name}} --fixed_grid --split_events \
                         --goes_position {slot[-4:]} --goes_sector conus --dx={dx} --dy={dy} {files_to_process}"
                    )

                if realtime:
                    if "ussamoa" in outroot:
                        os.system(
                            f"{ps_pythonpath}/bin/python {pltg}/lightningcast/glm/glmtools/examples/grid/make_GLM_grids.py \
                        -o {raw_outdir}/{{start_time:%Y/%b/%d}}/{{dataset_name}} --fixed_grid --split_events \
                        --goes_position {slot[-4:]} --dx={dx} --dy={dy} --ctr_lat -13 --ctr_lon -167.5 --width 1250 --height 900 {files_to_process}"
                        )
                    else:
                        os.system(
                            f"{ps_pythonpath}/bin/python {pltg}/lightningcast/glm/glmtools/examples/grid/make_GLM_grids.py \
                      -o {raw_outdir}/{{start_time:%Y/%b/%d}}/{{dataset_name}} --fixed_grid --split_events \
                      --goes_position {slot[-4:]} --goes_sector conus --dx={dx} --dy={dy} {files_to_process}"
                        )
                else:
                    new_files = []
                    tmpglm = os.environ["HOME"] + "/tmpglm/"
                    logging.info(f"copying files to {tmpglm}")
                    utils.mkdir_p(tmpglm)
                    for fff in glob(files_to_process):
                        shutil.copy(fff, tmpglm)
                        new_files.append(f"{tmpglm}/{os.path.basename(fff)}")
                    os.system(
                        f"{ps_pythonpath}/bin/python {pltg}/lightningcast/glm/glmtools/examples/grid/make_GLM_grids.py -o \
                    {raw_outdir}/{{start_time:%Y/%b/%d}}/{{dataset_name}} --fixed_grid --split_events --goes_position {slot[-4:]} \
                    --goes_sector conus --dx={dx} --dy={dy} {(' ').join(new_files)}"
                    )
                    for fff in new_files:
                        os.remove(fff)

                # remove the log
                try:
                    os.remove("make_GLM_grid.log")
                except (OSError, IOError) as err:
                    pass

                # make the aggregated grids and remapped grids
                # raw: OR_GLM-L2-GLMC-M3_G16_s20182411750000_e20182411751000_c20182411751360.nc
                rawfile = glob(
                    raw_outdir + oneMinAgo.strftime("/%Y/%b/%d/*s%Y%j%H%M*.nc")
                )
                if rawfile:

                    if realtime:
                        gzip = False
                    else:
                        gzip = False  # True
                        outroot = f"{outdir}/{oneMinAgo.strftime('%Y%m%d')}/GLM/{slot}/"

                    status = aggregate_glm_grids(
                        oneMinAgo.strftime("%Y%j%H%M"),
                        raw_outdir,
                        outdir=outroot,
                        gzip=False,
                        re_upload=re_upload,
                        geotiff=geotiffs,
                    )  # accumPeriod are default
                    # if(not(realtime)): gstat = call(['gzip','-f',rawfile[0]])

                    lastfile = latest_file

        if realtime:
            logging.info("sleeping...")
            time.sleep(5)
        else:
            dt += timedelta(minutes=1)  # increment if not realtime


if __name__ == "__main__":
    main()
