from datetime import datetime, timedelta
import pandas as pd
import scipy.ndimage
import glob
import os, sys, re
import collections
import pickle
from lightningcast import utils
from lightningcast.collect_data import read_netcdf
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.path as mpath
from matplotlib.ticker import MultipleLocator
import pyresample as pr
import logging
import netCDF4
from subprocess import PIPE, call, Popen

# from visualize_validation import verification
import pytz
import mysql.connector
import urllib

mytz = pytz.timezone("America/Chicago")  # For Grafana display of data

PW = "eEGOpgEuC8!_vQq]"
HOST = "db.ssec.wisc.edu"
DBNAME = "lightning_meteograms"  #'lightning_dss'
USERNAME = "lightning_meteograms"
cnx = 0  # initialization. Used as the mysql connection variable
# ------------------------------------------------------------------------------------------------
def execute_sql(sql_statement, return_query=True):
    global cnx

    cursor = cnx.cursor(dictionary=True)

    try:
        cursor.execute(sql_statement)
        if return_query:
            query_result = cursor.fetchall()
            cursor.close()
        else:
            cnx.commit()
            cursor.close()
    except mysql.connector.Error as err:
        logging.error(str(err))
        sys.exit(1)

    if return_query:
        return query_result


# ------------------------------------------------------------------------------------------------
def delete_db_tables(dbname=DBNAME):
    global cnx

    # connect to database
    try:
        cnx = mysql.connector.connect(
            user=USERNAME, password=PW, host=HOST, database=dbname
        )
    except mysql.connector.Error as err:
        logging.error(str(err))
        sys.exit(1)

    # Get tablenames
    sql = f"SHOW TABLES WHERE `Tables_in_{dbname}` NOT LIKE 'example'"
    result = execute_sql(sql)

    # Remove each table from DB
    for t in result:
        table = t[f"Tables_in_{dbname}"]
        sql = f"DROP TABLE `{dbname}`.`{table}`"
        execute_sql(sql, return_query=False)

        # sql = ("DELETE FROM `"+table+"` WHERE DATE_ADD(`time`,INTERVAL "+purge_thresh_days +" DAY) < UTC_TIMESTAMP() ")

        # status=delete_records(sql)


# ------------------------------------------------------------------------------------------------
def purge_records(purge_thresh_days, dbname=DBNAME):
    """
    Remove records in dbname that are older than purge_thresh_days.
    """
    global cnx

    # connect to database
    try:
        cnx = mysql.connector.connect(
            user=USERNAME, password=PW, host=HOST, database=dbname
        )
    except mysql.connector.Error as err:
        logging.error(str(err))
        sys.exit(1)

    # Get tablenames
    sql = f"SHOW TABLES WHERE `Tables_in_{dbname}` NOT LIKE 'example'"
    result = execute_sql(sql)

    # Remove old records from DB
    for t in result:
        table = t[f"Tables_in_{dbname}"]
        sql = f"DELETE FROM `{table}` WHERE DATE_ADD(`time`,INTERVAL {purge_thresh_days} DAY) < UTC_TIMESTAMP() "
        execute_sql(sql, return_query=False)


# ------------------------------------------------------------------------------------------------
def delete_one_table(table_name, dbname=DBNAME):
    """
    Removes <table_name> from <dbname>. Use with caution!
    """
    global cnx

    # connect to database
    try:
        cnx = mysql.connector.connect(
            user=USERNAME, password=PW, host=HOST, database=dbname
        )
    except mysql.connector.Error as err:
        logging.error(str(err))
        sys.exit(1)

    sql = f"DROP TABLE `{dbname}`.`{table_name}`"

    result = execute_sql(sql, return_query=False)


# ------------------------------------------------------------------------------------------------
def create_table(table_name):
    global cnx

    sql = "CREATE TABLE IF NOT EXISTS `%s` LIKE example" % table_name

    try:
        cursor = cnx.cursor()
        cursor.execute(sql)
        cnx.commit()
    except mysql.connector.Error as err:
        logging.error("could not create db table.")
        logging.error(str(err.msg))
        raise


# ------------------------------------------------------------------------------------------------
def insert_record(data):
    # data is a dictionary

    global cnx

    # get tablename
    location_name = data["location_name"]
    # We delete this key so now the dict is ready for upload
    del data["location_name"]

    # format time for MySQL datetime type: '%Y-%m-%d %H:%M:%S'

    # Call function to create DB table for this location
    try:
        create_table(location_name)
    except mysql.connector.Error as err:
        raise

    # Build sql insert statement pieces
    sql_fields = (
        "sat_name, meso_domain, utctime, localtime_ts, pltg_1fl_60min, pltg_1fl_60min_max16km, "
        + "glmct_8km_5min, glmct_16km_5min, enict_8km_5min, enict_16km_5min, cg_enict_8km_5min, cg_enict_16km_5min"
    )
    sql_values = (
        "%(sat_name)s, %(meso_domain)s, %(utctime)s, %(localtime_ts)s, %(pltg_1fl_60min)s, %(pltg_1fl_60min_max16km)s, "
        + "%(glmct_8km_5min)s, %(glmct_16km_5min)s, %(enict_8km_5min)s, %(enict_16km_5min)s, %(cg_enict_8km_5min)s, %(cg_enict_16km_5min)s"
    )

    sql = f"INSERT INTO `{location_name}` ({sql_fields}) VALUES({sql_values})"

    try:
        cursor = cnx.cursor()
        cursor.execute(sql, data)
        cnx.commit()
        cursor.close()
    except mysql.connector.Error as err:
        raise


# ------------------------------------------------------------------------------------------------
def insert_grafana_records(data_list, dbname=DBNAME):
    global cnx
    logging.info("Process started.")

    # connect to database
    try:
        cnx = mysql.connector.connect(
            user=USERNAME, password=PW, host=HOST, database=dbname
        )
    except mysql.connector.Error as err:
        logging.error(str(err.msg))
        raise

    # Upload data for each location contained in the list of dicts
    for v in data_list:
        try:
            result = insert_record(v)
        except mysql.connector.Error as err:
            raise

    cnx.close()

    logging.info("Process ended.")


# ----------------------------------------------------------------------------------------------------------------
from numba import jit


@jit(nopython=True)
def masked_median_filter_numba(image, mask, kernel_size):  # , padded_image):
    """
    Applies a median filter to specific regions of an image based on a mask using numba.

    Args:
      image: A 2D numpy array representing the image.
      mask: A 2D numpy array of the same size as the image, where True indicates
        regions to be filtered and False indicates regions to be left untouched.
      kernel_size: The size of the median filter kernel (default: 3).
      padded_image: Padded array to account for neighborhood size. Should be padded by <offset>.
                    np.pad is not supported by numba, so we need to supply this argument.

    Returns:
      A new 2D numpy array with the filtered image.
    """
    # Create a copy of the image to avoid modifying the original
    image_filtered = np.copy(image)

    # Calculate the offset for the kernel
    offset = kernel_size // 2

    ny, nx = image.shape

    padded_image = np.zeros((ny + (2 * offset), nx + (2 * offset)))
    padded_image[offset : ny + offset, offset : nx + offset] = image_filtered

    # Loop through each pixel
    for i in range(ny):
        for j in range(nx):
            # Check if the pixel is masked
            if mask[i, j]:
                # Define the neighborhood
                # We do "offset+ ..." because padded_image.shape != image.shape
                neighborhood = padded_image[
                    offset + i - offset : offset + i + offset + 1,
                    offset + j - offset : offset + j + offset + 1,
                ]
                # neighborhood = image[ i - offset: i + offset + 1, j - offset: j + offset + 1]

                # Apply the median filter to the neighborhood
                image_filtered[i, j] = np.median(neighborhood)

    return image_filtered


# ----------------------------------------------------------------------------------------------------------------
@jit(nopython=True)
def mask_data(data, badfillval, bad_data_fraction_threshold, hs, missingfillval):

    ny, nx = data.shape

    # Do our own padding, because numba doesn't like np.pad
    padded_data = np.zeros((ny + (2 * hs), nx + (2 * hs)))
    padded_data[hs : ny + hs, hs : nx + hs] = data

    new_data = np.copy(padded_data)  # copy of padded image data
    nny, nnx = new_data.shape

    mask = np.zeros(new_data.shape)

    neighborhood_size = (hs * 2) ** 2

    # Go through all pixels and determine DQF types
    for i in range(nny):
        for j in range(nnx):

            # That is, if it was masked. Note that we may be modifying new_data below
            if new_data[i, j] == badfillval:

                # Define the neighborhood
                neighborhood = padded_data[i - hs : i + hs + 1, j - hs : j + hs + 1]

                bad_data_fraction = (
                    np.count_nonzero(neighborhood == badfillval) / neighborhood_size
                )

                if bad_data_fraction > bad_data_fraction_threshold:
                    # Too much bad data in the neighborhood. Just fill with missingfillval.
                    new_data[i - hs : i + hs + 1, j - hs : j + hs + 1] = missingfillval
                    mask[i - hs : i + hs + 1, j - hs : j + hs + 1] = 1

                else:
                    # Mark regions to apply the median filter to a small neighborhood
                    mask[i - hs : i + hs + 1, j - hs : j + hs + 1] = 2

    # Subset out the actual new data
    new_data = new_data[hs : hs + ny, hs : hs + nx]
    mask = mask[hs : hs + ny, hs : hs + nx]

    return new_data, mask


# ----------------------------------------------------------------------------------------------------------------
def clean_data(sat_data, idict, ch):
    """
    Flags and attempts to smooth over bad data. E.g., missing scan line chunks and speckle noise.

    Args:
      sat_data: 2D numpy array
      idict: dict with a lot of config data and meta data and more-static data (e.g., idict['satzen']
      ch: string with channel name (e.g. 'CH02')

    Returns:
      sat_data or new_sat_data: new 2D numpy array with smoothed or masked regions, if necessary. Else, just the same array as sat_data, but filled (i.e., no longer a MaskedArray).

    """

    # Halfsize at 2-km res
    # This is the region we flag as potentially bad. And where we attempt to fix.
    hs = idict["qc_hs"]

    fillval = 4095
    mask = np.zeros(sat_data.shape, dtype=np.uint8)

    missingfillval = 0
    # meanfillval = idict['svals'][ch]['mean']

    check_05km_mask = False
    if ch == "CH02":
        # Fix the spacelook regions...still a masked array
        sat_data[idict["satzen_05km"] >= 90] = 0
        hs = hs * 4
        maskkey = "05km"
    elif ch == "CH05" or ch == "CH01" or ch == "CH03":
        # Fix the spacelook regions
        sat_data[idict["satzen_1km"] >= 90] = 0
        hs = hs * 2
        maskkey = "1km"
        check_05km_mask = True
    else:
        # Fix the spacelook regions
        sat_data[idict["satzen"] >= 90] = 0
        maskkey = "2km"
        if maskkey not in idict["masks"]:
            idict["masks"][
                maskkey
            ] = mask  # We always have at least a 2km mask for the output netcdf

    try:
        bad_data = np.count_nonzero(sat_data.mask)
    except AttributeError as err:
        # This means sat_data wasn't passed in as a MaskedArray. Likely Himawari data.
        sat_data = np.ma.MaskedArray(
            sat_data, mask=np.isnan(sat_data), fill_value=fillval
        )
        bad_data = np.count_nonzero(sat_data.mask)

    # Mask out data if we find some bad/masked data that is in an area of good satzen
    if bad_data > 0:
        logging.warning(f"Found bad data in channel: {ch}")

        sat_data = sat_data.filled(
            fillval
        )  # make it a normal np.ndarray (i.e., not masked array)

        bad_data_fraction_threshold = idict["bad_data_fraction_threshold"]

        # Numba-enhanced function
        new_sat_data, mask = mask_data(
            sat_data,
            fillval,
            bad_data_fraction_threshold,
            hs,
            missingfillval=missingfillval,
        )

        idict["masks"][maskkey] = mask  # save the mask

        # fig,ax = plt.subplots(nrows=1,ncols=2)
        # ax[0].imshow(new_sat_data, vmin=0,vmax=1)
        # ax[1].imshow(mask)
        # plt.show()
        # sys.exit()

        if np.max(mask) == 2:
            # Numba-enhanced function
            new_sat_data = masked_median_filter_numba(new_sat_data, mask, 7)

        # Check the 05 km mask, if reflective band. We want to mask all reflective bands in the
        # same fashion. This produces better results. Note that if 0.5 km data, we won't be able
        # to get the 1 km mask.
        if check_05km_mask:
            if "05km" in idict["masks"]:
                # For "large" areas of masked pixels in 0.5km mask, apply also to 1km reflective band data.
                # These are regions where mask == 1.
                mask1km = np.ma.masked_equal(idict["masks"]["05km"][::2, ::2], 1).mask
                new_sat_data = np.ma.masked_array(new_sat_data, mask=mask1km)
                return new_sat_data.filled(missingfillval)
            else:
                return new_sat_data
        else:
            return new_sat_data

    else:  # No bad pixels in sat_data channel

        # Check the 05 km mask, if reflective band. We want to mask all reflective bands in the
        # same fashion. This produces better results. Note that if 0.5 km data, we won't be able
        # to get the 1 km mask.
        if check_05km_mask:
            if "05km" in idict["masks"]:
                # For "large" areas of masked pixels in 0.5km mask, apply also to 1km reflecitve band data.
                # These are regions where mask == 1.
                mask1km = np.ma.masked_equal(idict["masks"]["05km"][::2, ::2], 1).mask
                sat_data = np.ma.masked_array(sat_data, mask=mask1km)

        return sat_data.filled(missingfillval)


# ----------------------------------------------------------------------------------------------------------------
class location:
    def __init__(self, code, longname, latlon, city, country):
        self.code = code
        self.longname = longname
        self.latlon = latlon
        self.city = city
        self.country = country


# ----------------------------------------------------------------------------------------------------------------
def get_locations(locations_file):
    locations = []

    of = open(
        locations_file, "r"
    )  # from https://www.partow.net/miscellaneous/airportdatabase/
    lines = of.readlines()
    of.close()

    if os.path.basename(locations_file).startswith("stadiums"):
        for line in lines[1:]:  # skip header
            parts = line.split(",")
            locations.append(
                location(
                    code=parts[3],
                    longname=parts[0],
                    latlon=(float(parts[-2]), float(parts[-1])),
                    city=parts[1],
                    country="USA",
                )
            )
            # Camp Randall Stadium,Madison,WI,Wisconsin,Big Ten,80321,1917,2005,fbs,43.06993802646878,-89.41274642944336
    else:
        for line in lines:
            if (
                line.startswith("K")
                or line.startswith("C")  # USA48
                or line.startswith("P")  # Canada
                or line.startswith("T")  # Hawaii/Alaska
                or line.startswith("M")  # Puerto Rico, US Virgin Islands, et al.
                or line.startswith("S")  # Mexico and Central America
            ):  # South America
                parts = line.split(":")
                if float(parts[-2]) == 0 and float(parts[-1]) == 0:
                    continue  # lat and lon are 0
                locations.append(
                    location(
                        code=parts[0],
                        longname=parts[2],
                        latlon=(float(parts[-2]), float(parts[-1])),
                        city=parts[3],
                        country=parts[4],
                    )
                )
                # KABI:ABI:ABILENE RGNL:ABILENE:USA:032:024:040:N:099:040:054:W:00546:32.411:-99.682

    return locations


# ----------------------------------------------------------------------------------------------------------------
def get_ondemand_locations(remote_csv=os.environ["ONDEMAND_CSV"]):
    logging.info("Process started")
    now = datetime.utcnow()

    ondemand_locations = []

    try:
        candidate_sites = pd.read_csv(remote_csv)
    except urllib.error.HTTPError as err:
        logging.error(str(err))
        logging.error("Returning empty list of ondemand_locations")
        return ondemand_locations

    num_entries = len(candidate_sites)
    if num_entries > 2000:
        logging.warning(f"Too many ondemand sites ({num_entries}). Skipping.")
        return

    # QC and select the locations
    # Make lists. Colnames are hardcoded.
    try:
        ts_of_request = list(candidate_sites["Timestamp"])  # type=str
        emails = list(candidate_sites["Email Address"])  # type=str
        event_name = list(candidate_sites["Unique name of event"])  # type=str
        lats = list(candidate_sites["Latitude of event (degrees north)"])  # type=float
        lons = list(candidate_sites["Longitude of event (degrees west)"])  # type=float
        begin_event_ts = list(
            candidate_sites["Beginning date & time of request (UTC)"]
        )  # type=str
        end_event_ts = list(
            candidate_sites["End date & time of request (UTC)"]
        )  # type=str
    except KeyError:
        logging.error("Bad keys. Returning empty list of ondemand_locations.")
        return ondemand_locations
    else:
        # These checks are redundant with the javascript checks in the Apps Script.
        # We do these here in case the program receives a request before the Apps
        # Script can delete the row.
        for ii in range(len(emails)):

            # Check for valid email
            if not (
                bool(re.match(r"^(probsevere.*|.*noaa\.gov|.*wisc\.edu)+$", emails[ii]))
            ):
                continue

            # Check for redundant event name that occurs later
            if event_name[ii] in event_name[ii + 1 :]:
                continue

            # Check event_name - must be alphanumeric, underscores, or hyphens only
            if not (bool(re.match(r"^[A-Za-z0-9_\- ]+$", event_name[ii]))):
                continue

            begin_dt = datetime.strptime(
                begin_event_ts[ii], "%m/%d/%Y %H:%M:%S"
            )  # assumed to be UTC
            end_dt = datetime.strptime(
                end_event_ts[ii], "%m/%d/%Y %H:%M:%S"
            )  # assumed to be UTC

            # Check that end_dt is after begin_dt
            # Also check that duration is within bounds.
            if end_dt <= begin_dt or (end_dt - begin_dt).seconds / 86400 > float(
                os.environ["DSS_DURATION_THRESH"]
            ):
                continue  # 7-day duration limit

            if (
                begin_dt <= now <= end_dt
            ):  # Implicitly checks that it's not an "expired" event.
                lon = lons[ii] * -1 if (lons[ii] > 0) else lons[ii]
                # Make location object
                ondemand_locations.append(
                    location(
                        code="DSS",
                        longname=event_name[ii],
                        latlon=((lats[ii], lon)),
                        city="N/A",
                        country="N/A",
                    )
                )

        logging.info("Process ended")
        return ondemand_locations


# ----------------------------------------------------------------------------------------------------------------
def get_L2_flashes(glm_L2_patt, start_dt, end_dt):
    """
    Uses the glm_L2_patt and start_dt and end_dt to get GLM L2 flashes.
    The program will gather flash lat/lon data from startdt to end_dt, iterating by 1 min.
    Returns a dictionary of GLM-deteced flash lats/lons between start_dt and end_dt.
    """

    logging.info("Process started.")

    assert start_dt <= end_dt
    satlon = -999

    L2_flashes = {"lats": [], "lons": []}

    dt = start_dt
    while dt <= end_dt:
        files = np.sort(glob.glob(dt.strftime(glm_L2_patt)))
        for f in files:  # is usu. 3 files; every 20 sec for 1 min
            logging.info(f"Reading {os.path.basename(f)}")
            try:
                nc = netCDF4.Dataset(f, "r")
                L2_flashes["lats"] += list(nc.variables["flash_lat"][:])
                L2_flashes["lons"] += list(nc.variables["flash_lon"][:])
                if satlon == -999:  # needed for plax correction
                    satlon = nc.variables["nominal_satellite_subpoint_lon"][:]
                nc.close()
            except (OSError, IOError, FileNotFoundError) as err:
                logging.error(str(err))
        dt = dt + timedelta(minutes=1)

    # L2 flashes are already parallax corrected!

    # Parallax correct lats and lons
    # if plax_corr_hgt > 0 and len(L2_flashes["lats"]) > 0:
    #    logging.info(
    #        f"Parallax-correcting flash lats and lons using constant parallax height={plax_corr_hgt}m"
    #    )
    #    L2_flashes["lats"], L2_flashes["lons"] = utils.plax(
    #        lats=np.array(L2_flashes["lats"]),
    #        lons=np.array(L2_flashes["lons"]),
    #        satlon=satlon,
    #        elevation=plax_corr_hgt,
    #    )

    L2_flashes["lats"] = np.array(L2_flashes["lats"])
    L2_flashes["lons"] = np.array(L2_flashes["lons"])

    logging.info("Process ended.")
    return L2_flashes


# ----------------------------------------------------------------------------------------------------------------
def floor_dt(dt, delta):
    assert isinstance(delta, timedelta)
    return datetime.min + np.floor((dt - datetime.min) / delta) * delta


# ----------------------------------------------------------------------------------------------------------------
def get_eni_flashes(eni_patt, start_dt, end_dt):
    """
    Uses the eni_patt and start_dt and end_dt to get ENI flashes.
    The program will gather flash lat/lon data from startdt to end_dt, iterating by 1 min.
    Returns a dictionary of ENI-deteced flash lats/lons/flash types between start_dt and end_dt.
    """

    logging.info("Process started.")

    assert start_dt <= end_dt

    eni_flashes = {"lats": [], "lons": [], "types": []}
    dt = start_dt

    # Determine which type of ENI data we're dealing with (realtime or historical), and process accordingly
    if os.path.basename(eni_patt).startswith(
        "flash"
    ):  # realtime. startswith('entln') is historical data

        while dt <= end_dt:
            files = np.sort(
                glob.glob(dt.strftime(eni_patt))
            )  # Should be at most 1 file per minute
            if len(files) > 0:
                thefile = files[0]
                logging.info(f"Reading {os.path.basename(thefile)}")
                try:
                    df = pd.read_csv(thefile)
                    eni_flashes["lats"] += list(df["Latitude"])
                    eni_flashes["lons"] += list(df["Longitude"])
                    eni_flashes["types"] += list(df["Flash_Type"])
                except (OSError, IOError, FileNotFoundError) as err:
                    logging.error(str(err))
            dt += timedelta(minutes=1)

    else:  # historical E.g., "/ships19/grain/lightning/%Y/%Y_%m_%d_%j/%H/entln_%Y%j_%H%M*.txt"

        # Floor the *end*_dt to the nearest 5 min. In these files, the timestamp is the approximate beginning time of data in the 5-min file
        nearest_floor_5min_dt = floor_dt(end_dt, timedelta(minutes=5))
        prev_5min_dt = nearest_floor_5min_dt - timedelta(minutes=5)
        files = glob.glob(nearest_floor_5min_dt.strftime(eni_patt)) + glob.glob(
            prev_5min_dt.strftime(eni_patt)
        )
        for f in files:
            logging.info(f"Reading {f}")
            of = open(f, "r")
            lines = of.readlines()
            of.close()
            for line in lines:
                # e.g., '"time":"2023-02-19T21:41:08.599084335Z","type":0,"latitude":-14.42152,"longitude":29.06414, ......
                parts = line.split(",")
                test_dt = datetime.strptime(
                    parts[0].split('":"')[1][0:19], "%Y-%m-%dT%H:%M:%S"
                )
                if start_dt <= test_dt <= end_dt:
                    eni_flashes["types"].append(int(parts[1].split(":")[1]))
                    eni_flashes["lats"].append(float(parts[2].split(":")[1]))
                    eni_flashes["lons"].append(float(parts[3].split(":")[1]))

    eni_flashes["types"] = np.array(eni_flashes["types"])
    eni_flashes["lats"] = np.array(eni_flashes["lats"])
    eni_flashes["lons"] = np.array(eni_flashes["lons"])

    logging.info("Process ended.")
    return eni_flashes


# ----------------------------------------------------------------------------------------------------------------
def lightning_meteograms_inmem(
    prob_grid,
    irx,
    iry,
    dt,
    projection,
    locations_file,
    glm_L2_patt,
    meta,
    ondemand_sites=False,
    eni_patt=None,
    code_includes=[],
    plax_hgt=0,
):
    """
    This function returns a list dictionaries with probability info and lightning info for select
    locations. Each dict is for a given location at a given time.

    prob_grid: 2D numpy array with shape = (ny,nx). This is the probabilities (0-1) for one time.
    irx: 1D numpy array with len = nx. These are the projection-x coords
    iry: 1D numpy array with len = ny. These are the projection-y coords
    dt: datetime.datetime for the data in prob_grid. It is typically the satellite scan time.
    projection: pyproj.Proj object that describes projection of prob_grid
    locations_file: <string> Filename that contains the names and lat/lons of every location we want to collect data for.
    glm_L2_patt: <string> Pattern of where to find the glm_L2 data. E.g. /arcdata/goes/grb/goes16/%Y/%Y_%m_%d_%j/glm/L2/LCFA/*_s%Y%j%H%M*
    meta: <dict> with scene-specific info
    ondemand_sites: True/False flag to process additional on-demand sites for LC meteograms. Uses on.environ["ONDEMAND_CSV"].
    eni_patt: <string> (or None) Pattern of where to find the ENI flash CSVs. E.g., /data/PLTG/ENI/realtime/flash-%Y-%m-%dT%H-%M-.csv
    code_includes: <list> of string patterns. Only use locations that begin with these patterns.
    plax_hgt: Height in meters used to parallax-correct GLM flash points. Default=0 (no correction).
    """

    logging.info("Process started")

    try:
        assert prob_grid.shape == (len(iry), len(irx))
    except AssertionError:
        raise

    sat_name = f"{meta['platform_name']} {meta['sensor'].upper()}"
    meso_domain = "1" if (meta["scene_abbr"].startswith("M")) else "0"

    locations = get_locations(locations_file)  # a dict of names and lat/lon tuples

    # Get on-demand locations and add to locations
    if ondemand_sites:
        locations += get_ondemand_locations()

    if len(code_includes) > 0:
        new_locations = []
        for code in code_includes:
            new_locs = [nloc for nloc in locations if nloc.code.startswith(code)]
            new_locations += new_locs
        locations = new_locations.copy()

    timestamp = dt.strftime("%Y-%m-%d %H:%M:%S")  # MySQL-formatted timestamp (UTC)
    mytz_ts = dt.astimezone(mytz).strftime(
        "%Y-%m-%d %H:%M:%S"
    )  # MySQL-formatted timestamp (local time). Needed for proper Grafana display.

    # Get L2 GLM flashes
    glm_flashes = get_L2_flashes(
        glm_L2_patt,
        start_dt=dt - timedelta(minutes=5),
        end_dt=dt,
    )

    # Get ENI flashes
    if eni_patt:
        eni_flashes = get_eni_flashes(
            eni_patt, start_dt=dt - timedelta(minutes=5), end_dt=dt
        )
    else:
        eni_flashes = {"lats": []}

    lastx = len(irx) - 1
    lasty = len(iry) - 1

    loc_dicts = []
    for loc in locations:
        lat, lon = loc.latlon
        x, y = projection(lon, lat)
        xidx, _ = utils.find_nearest(irx, x)
        yidx, _ = utils.find_nearest(iry, y)

        # If the nearest proj coords for the location are within the grid (i.e., not on the edge), then continue to collect data
        if xidx > 0 and xidx < lastx and yidx > 0 and yidx < lasty:
            location_name = (
                loc.longname if (loc.code == "") else f"{loc.code}: {loc.longname}"
            )
            newdict = {
                "sat_name": sat_name,
                "meso_domain": meso_domain,
                "utctime": timestamp,
                "location_name": location_name,
                "glmct_8km_5min": None,
                "glmct_16km_5min": None,
                "enict_8km_5min": None,
                "enict_16km_5min": None,
                "cg_enict_8km_5min": None,
                "cg_enict_16km_5min": None,
                "localtime_ts": mytz_ts,
            }
            newdict["pltg_1fl_60min"] = str(round(prob_grid[yidx, xidx] * 100))
            try:
                newdict["pltg_1fl_60min_max16km"] = str(
                    round(
                        np.max(prob_grid[yidx - 8 : yidx + 9, xidx - 8 : xidx + 9])
                        * 100
                    )
                )  # 8 2-km pixels
            except ValueError:
                newdict["pltg_1fl_60min_max16km"] = newdict[
                    "pltg_1fl_60min"
                ]  # just duplicate

            # Now collect the L2 GLM flashes
            # Get count of nearby flashes (points). Assumes 0.01 = 1km
            newdict["glmct_8km_5min"] = str(
                np.count_nonzero(
                    (glm_flashes["lats"] >= (lat - 0.08))
                    & (glm_flashes["lats"] <= (lat + 0.08))
                    & (glm_flashes["lons"] >= (lon - 0.08))
                    & (glm_flashes["lons"] <= (lon + 0.08))
                )
            )
            newdict["glmct_16km_5min"] = str(
                np.count_nonzero(
                    (glm_flashes["lats"] >= (lat - 0.16))
                    & (glm_flashes["lats"] <= (lat + 0.16))
                    & (glm_flashes["lons"] >= (lon - 0.16))
                    & (glm_flashes["lons"] <= (lon + 0.16))
                )
            )

            # Now collect ENI flashes (points)
            if len(eni_flashes["lats"]) > 0:
                newdict["enict_8km_5min"] = str(
                    np.count_nonzero(
                        (eni_flashes["lats"] >= (lat - 0.08))
                        & (eni_flashes["lats"] <= (lat + 0.08))
                        & (eni_flashes["lons"] >= (lon - 0.08))
                        & (eni_flashes["lons"] <= (lon + 0.08))
                    )
                )
                newdict["enict_16km_5min"] = str(
                    np.count_nonzero(
                        (eni_flashes["lats"] >= (lat - 0.16))
                        & (eni_flashes["lats"] <= (lat + 0.16))
                        & (eni_flashes["lons"] >= (lon - 0.16))
                        & (eni_flashes["lons"] <= (lon + 0.16))
                    )
                )
                newdict["cg_enict_8km_5min"] = str(
                    np.count_nonzero(
                        (eni_flashes["lats"] >= (lat - 0.08))
                        & (eni_flashes["lats"] <= (lat + 0.08))
                        & (eni_flashes["lons"] >= (lon - 0.08))
                        & (eni_flashes["lons"] <= (lon + 0.08))
                        & (eni_flashes["types"] == 1)
                    )
                )  # type == 1 --> CG
                newdict["cg_enict_16km_5min"] = str(
                    np.count_nonzero(
                        (eni_flashes["lats"] >= (lat - 0.16))
                        & (eni_flashes["lats"] <= (lat + 0.16))
                        & (eni_flashes["lons"] >= (lon - 0.16))
                        & (eni_flashes["lons"] <= (lon + 0.16))
                        & (eni_flashes["types"] == 1)
                    )
                )  # type == 1 --> CG

            loc_dicts.append(newdict)

    logging.info("Process ended")

    return loc_dicts


# ----------------------------------------------------------------------------------------------------------------
def lightning_meteograms_from_netcdf(
    lc_netcdf, locations_file, path_of_csvs, glm_netcdf=None
):
    logging.info("Process started")
    logging.error("This function is not operational at this time.")
    return

    hr_threshold = 2
    # make outdir
    utils.mkdir_p(path_of_csvs)

    locations = get_locations(locations_file)  # a dict of names and lat/lon tuples

    # read LightningCast netcdf
    lcvar = "PLTG_60min_plax"
    lcdata = {lcvar: {"atts": []}}
    global_atts = {
        "satellite": -1,
        "time_coverage_start": -1,
        "NWlat": -1,
        "NWlon": -1,
        "dx": -1,
        "dy": -1,
        "sector": -1,
    }
    utils.read_netcdf(lc_netcdf, datasets=lcdata, global_atts=global_atts)
    # make lats/lons
    nlats, nlons = lcdata[lcvar]["data"].shape
    SWlat = global_atts["NWlat"] - global_atts["dy"] * nlats
    lats = np.arange(global_atts["NWlat"], SWlat, -global_atts["dy"])
    NElon = global_atts["NWlon"] + global_atts["dx"] * nlons
    lons = np.arange(global_atts["NWlon"], NElon, global_atts["dx"])
    # format dt
    # 20230203-154617 UTC --> 2023/02/03 15:46:17
    curr_dt = datetime.strptime(global_atts["time_coverage_start"], "%Y%m%d-%H%M%S UTC")
    timestamp = curr_dt.strftime("%Y/%m/%d %H:%M:%S")

    # read_glm_netcdf #TODO

    for loc in locations:
        lat, lon = loc.latlon
        if lat == 0 and lon == 0:
            continue
        # find data indices of lat/lon point, within ltg_netcdf
        latidx, _ = utils.find_nearest(lats, lat)
        lonidx, _ = utils.find_nearest(lons, lon)
        pltg = lcdata[lcvar]["data"][latidx, lonidx]
        pltg_20km = np.max(
            lcdata[lcvar]["data"][latidx - 2 : latidx + 3, lonidx - 2 : lonidx + 3]
        )
        # open csv
        try:
            of = open(f"{path_of_csvs}/{loc.code}.csv", "r")
        except FileNotFoundError:
            of = open(f"{path_of_csvs}/{loc.code}.csv", "w")
            of.write("Date/Time (UTC),P(LTG),max[P(LTG w/in 20km)]\n")
        else:
            lines = of.readlines()
            of.close()
            # remove "old" times while writing out good times.
            of = open(f"{path_of_csvs}/{loc.code}.csv", "w")
            for ll, line in enumerate(lines):
                if line.startswith("Date"):
                    of.write(line)
                else:
                    dt = datetime.strptime(line.split(",")[0], "%Y/%m/%d %H:%M:%S")
                    if (curr_dt - dt).seconds / 3600 <= hr_threshold:
                        of.write(line)
        # Now write the latest time
        of.write((",").join([timestamp, f"{int(pltg)}", f"{int(pltg_20km)}"]) + "\n")
        of.close()

    logging.info("Process ended")


#################################################################################################
def make_dbase(data, pixbuff=5, stride=40):
    ltg_dbase = {}
    ny, nx = data.shape
    for yy in range(pixbuff, ny - pixbuff, stride):
        for xx in range(pixbuff, nx - pixbuff, stride):
            ltg_dbase[(yy, xx)] = {"pltg": [], "dts": [], "max_fed": [], "solzen": []}

    return ltg_dbase


#################################################################################################i
def get_fed(
    dt,
    glmpatt,
    glmvar="flash_extent_density",
    glm_area_def=None,
    target_area_def=None,
    remap_cache=None,
    radius_of_influence=4000,
    fill_value=-1,
):

    if glmpatt.startswith("http"):  # signifies UMD FD file
        tmpfile = "tmpGLMfile.nc"
        minute_offsets = [0, 1, 2, -1, -2]
        for mmm in minute_offsets:
            tmpdt = dt - timedelta(mmm)
            remotefilename = tmpdt.strftime(glmpatt)
            proc = Popen(
                [
                    "wget",
                    remotefilename,
                    "-O",
                    tmpfile,
                    "-t",
                    "1",
                    "--no-check-certificate",
                ],
                stdout=PIPE,
                stderr=PIPE,
            )
            stdout, stderr = proc.communicate()
            if not (isinstance(stdout, str)):
                stdout = stdout.decode("utf-8")
            if not (isinstance(stderr, str)):
                stderr = stderr.decode("utf-8")
            if "not found" in stderr.lower():
                logging.error(f"{remotefilename} not found.")
                if mmm == minute_offsets[-1]:
                    logging.warning(f"Can't find any GLM files near {dt}")
                    raise ValueError
            elif os.path.isfile(tmpfile):  # first check if it exists
                if (
                    os.path.getsize(tmpfile) > 1024
                ):  # if not at least 1kB, then something is wrong.
                    logging.info(f"Got {remotefilename}")
                    break
                else:
                    # file size is probably 0
                    logging.warning(f"File size is likely 0. Removing {tmpfile}.")
                    try:
                        os.remove(outfilename)
                    except:
                        pass

        nc = netCDF4.Dataset(tmpfile, "r")
        fed = nc.variables["Flash_extent_density_window"][
            :
        ]  # hard-coded for now #FIXME
        nc.close()

    else:
        # glmfile = glob.glob(dt.strftime(os.path.join(glmdir,'%Y%m%d','GLM',sat,'agg','%Y%m%d-%H%M*.netcdf*')))
        glmfile = glob.glob(dt.strftime(glmpatt))
        if not (glmfile):
            minute_offsets = [-1, 1, -2, 2, -3, 3]
            for mmm in minute_offsets:
                newdt = dt - timedelta(minutes=mmm)
                glmfile = glob.glob(
                    newdt.strftime(glmpatt)
                )  # os.path.join(glmdir,'%Y%m%d','GLM',sat,'agg','%Y%m%d-%H%M*.netcdf*')))
                if glmfile:
                    break
        if glmfile:
            glmdata = {glmvar: -1}
            ss = read_netcdf(glmfile[0], data=glmdata)
            fed = glmdata[glmvar][:]

    if remap_cache:
        try:
            remapped_fed = pr.kd_tree.get_sample_from_neighbour_info(
                "nn",
                remap_cache["target_def"].shape,
                fed,
                remap_cache["input_index"],
                remap_cache["output_index"],
                remap_cache["index_array"],
            )
        except ValueError as err:
            logging.warning(str(err))
            logging.warning("Most likely using older GLM dims. Attempting fix...")
            padded_fed = np.pad(
                fed, ((0, 1), (0, 1)), mode="constant", constant_values=0
            )  # padding right and bottom edges with zeros.
            try:
                remapped_fed = pr.kd_tree.get_sample_from_neighbour_info(
                    "nn",
                    remap_cache["target_def"].shape,
                    padded_fed,
                    remap_cache["input_index"],
                    remap_cache["output_index"],
                    remap_cache["index_array"],
                )
            except ValueError as err:
                raise
            else:
                return remapped_fed
        else:
            return remapped_fed

    elif glm_area_def and target_area_def:
        try:
            remapped_fed = pr.kd_tree.resample_nearest(
                glm_area_def,
                fed,
                target_area_def,
                radius_of_influence=radius_of_influence,
                fill_value=-1,
            )
        except UnboundLocalError:
            raise
        else:
            return remapped_fed

    else:
        try:
            return fed
        except UnboundLocalError:
            raise  # logging.critical("Can't remap GLM data");sys.exit()


#################################################################################################
def update_ltg_info(
    ltg_dbase, new_preds, glmdir, dt, solzen, fed=None, pixbuff=5, sat="goes_east"
):

    if fed is None:  # get the data #FIXME won't work...fed should always be present
        # get glm data
        glmfile = glob.glob(
            dt.strftime(
                os.path.join(
                    glmdir, "%Y%m%d", "GLM", sat, "agg", "%Y%m%d-%H%M*.netcdf*"
                )
            )
        )
        if not (glmfile):
            for mm in [1, 2, 3, 4, 5]:
                newdt = dt - timedelta(minutes=mm)
                glmfile = glob.glob(
                    newdt.strftime(
                        os.path.join(
                            glmdir, "%Y%m%d", "GLM", sat, "agg", "%Y%m%d-%H%M*.netcdf*"
                        )
                    )
                )
                if glmfile:
                    break
        if glmfile:
            glmdata = {"flash_extent_density": -1}
            ss = read_netcdf(glmfile[0], data=glmdata)
            fed = glmdata["flash_extent_density"][glmstartY:glmendY, glmstartX:glmendX]

    assert fed.shape == new_preds.shape

    # go through each point
    for (
        pt
    ) in (
        ltg_dbase
    ):  # pt is a two-element tuple with i,j coords in data space of our domain
        ltg_dbase[pt]["dts"].append(dt)
        ltg_dbase[pt]["pltg"].append(
            np.max(
                new_preds[
                    -pixbuff + pt[0] : pt[0] + pixbuff + 1,
                    -pixbuff + pt[1] : pt[1] + pixbuff + 1,
                ]
            )
        )
        maxfed = np.max(
            fed[
                -pixbuff + pt[0] : pt[0] + pixbuff + 1,
                -pixbuff + pt[1] : pt[1] + pixbuff + 1,
            ]
        )
        if isinstance(maxfed, np.ma.core.MaskedConstant):
            maxfed = 0
        ltg_dbase[pt]["max_fed"].append(maxfed)
        ltg_dbase[pt]["solzen"].append(solzen[pt])  # just using center point

    # ltg_dbase will be returned modified


#################################################################################################
def find_indices(lst, condition):
    return [i for i, elem in enumerate(lst) if condition(elem)]


#################################################################################################
def score_ltg_pts(ltg_dbase):

    fed_thresh = 0.1

    thresholds = [0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    hits, misses, FAs, CNs = [np.zeros(len(thresholds)) for _ in range(4)]
    POD, FAR, CSI, med_lts = [np.zeros(len(thresholds)) for _ in range(4)]

    leadtimes = collections.OrderedDict()
    for t in thresholds:
        leadtimes[t] = []
    spatial_leadtimes = {}

    nprev_times = 7  # number of previous times with no ltg obs. 7 = 30 min

    for pt in ltg_dbase:  # for each (i,j) point
        start_time = ltg_dbase[pt]["dts"][0]
        end_time = ltg_dbase[pt]["dts"][-1]
        for tt in range(
            nprev_times, len(ltg_dbase[pt]["dts"]) - 12
        ):  # for each scan time; #hardcoded -12 (-60 min...) for now
            # if we have met the minimum pltg thresh and don't have ltg at the current time,
            # then score the 'forecast' and measure lead-time, if applicable
            if (
                np.max(ltg_dbase[pt]["max_fed"][tt - nprev_times : tt + 1]) == 0
                and ltg_dbase[pt]["pltg"][tt] >= thresholds[0]
            ):
                # find tt+60 min
                # find max fed within tt+60 min
                max_fed_60 = np.max(
                    ltg_dbase[pt]["max_fed"][tt : tt + 12]
                )  # hardcoded 12*5 min
                for k, th in enumerate(thresholds):
                    if max_fed_60 >= fed_thresh:
                        if ltg_dbase[pt]["pltg"][tt] >= thresholds[k]:
                            hits[k] += 1
                            first_ltg_ind = (
                                tt
                                + find_indices(
                                    ltg_dbase[pt]["max_fed"][tt : tt + 12],
                                    lambda e: e >= fed_thresh,
                                )[0]
                            )
                            # 2.5 min offset since we don't know if ltg came at beginning or end of next 5 min interval,
                            # so we're approximating it by using the average.
                            LT = (
                                ltg_dbase[pt]["dts"][first_ltg_ind]
                                - ltg_dbase[pt]["dts"][tt]
                            ).seconds / 60 - 2.5
                            leadtimes[th].append(LT)
                            # populate spatial leadtimes dict
                            try:
                                spatial_leadtimes[pt][th].append(LT)
                            except KeyError:
                                spatial_leadtimes[pt] = collections.OrderedDict()
                                # create all thresholds
                                for t in thresholds:
                                    spatial_leadtimes[pt][t] = []
                                # now add the LT
                                spatial_leadtimes[pt][th].append(LT)

                        else:
                            misses[k] += 1
                    else:
                        if ltg_dbase[pt]["pltg"][tt] >= thresholds[k]:
                            FAs[k] += 1
                        else:
                            CNs[k] += 1

    for k, th in enumerate(thresholds):
        POD[k] = hits[k] / (hits[k] + misses[k])
        FAR[k] = FAs[k] / (hits[k] + FAs[k])
        CSI[k] = hits[k] / (hits[k] + misses[k] + FAs[k])
        med_lts[k] = np.median(leadtimes[th])

    return {
        "thresholds": thresholds,
        "hits": hits,
        "misses": misses,
        "FAs": FAs,
        "leadtimes": leadtimes,
        "spatial_leadtimes": spatial_leadtimes,
        "CNs": CNs,
        "POD": POD,
        "FAR": FAR,
        "CSI": CSI,
        "med_lts": med_lts,
        "num_pts": len(ltg_dbase),
        "start_time": start_time,
        "end_time": end_time,
    }


#################################################################################################
def aggregate_raw_stats_and_eval(
    list_of_pkls, outdir, fed_thresh=0.1, solzen_thresh=88
):
    """
    This is typically used to combine multiple ltg_dbase.pkls, like
    if you ran one per day for a month. We do this to get combined
    stats and run it through the verification function, often for
    a test dataset that can be broken up by solzen threshold.
    """

    preds = []
    target = []
    solzen = []
    for p, pkl in enumerate(list_of_pkls):
        print(pkl)
        dbase = pickle.load(open(pkl, "rb"))
        for pt, info in dbase.items():
            for ii in range(
                len(info["pltg"]) - 12
            ):  # hardcoded 12 for 12*5 min = 60min
                preds.append(info["pltg"][ii])
                solzen.append(info["solzen"][ii])
                if np.max(info["max_fed"][ii : ii + 12]) > fed_thresh:
                    target.append(1)
                else:
                    target.append(0)

    # evaluation
    preds = np.array(preds)
    target = np.array(target)
    solzen = np.array(solzen)

    # DAY
    print("DAYtime evaluation...")
    good_ind = np.where(solzen <= solzen_thresh)
    finaloutdir = os.path.join(outdir, "DAY")
    utils.mkdir_p(finaloutdir)
    scores_dict = verification(preds[good_ind], target[good_ind], finaloutdir)
    pickle.dump(scores_dict, open(os.path.join(finaloutdir, "scores_dict.pkl"), "wb"))
    stats = {
        "conv_preds": preds[good_ind],
        "labels": target[good_ind],
        "solzen": solzen[good_ind],
    }
    pickle.dump(stats, open(os.path.join(finaloutdir, "preds_labs.pkl"), "wb"))

    # NIGHT
    print("NIGHTtime evaluation...")
    good_ind = np.where(solzen > solzen_thresh)
    finaloutdir = os.path.join(outdir, "NIGHT")
    utils.mkdir_p(finaloutdir)
    scores_dict = verification(preds[good_ind], target[good_ind], finaloutdir)
    pickle.dump(scores_dict, open(os.path.join(finaloutdir, "scores_dict.pkl"), "wb"))
    stats = {
        "conv_preds": preds[good_ind],
        "labels": target[good_ind],
        "solzen": solzen[good_ind],
    }
    pickle.dump(stats, open(os.path.join(finaloutdir, "preds_labs.pkl"), "wb"))

    # ALL
    print("ALLtime evaluation...")
    finaloutdir = os.path.join(outdir, "ALL")
    utils.mkdir_p(finaloutdir)
    scores_dict = verification(preds, target, finaloutdir)
    pickle.dump(scores_dict, open(os.path.join(finaloutdir, "scores_dict.pkl"), "wb"))
    stats = {"conv_preds": preds, "labels": target, "solzen": solzen}
    pickle.dump(stats, open(os.path.join(finaloutdir, "preds_labs.pkl"), "wb"))


#################################################################################################
def combine_ltg_preds_and_labs(list_of_pickles, outdir):
    """
    Open and combine all_preds_labs.pkl files. For making a combine pikle for
    use in visualize_validation.py.
    """

    for i, pkl in enumerate(list_of_pickles):
        print(pkl)
        if i == 0:
            all_preds_labs = pickle.load(open(pkl, "rb"))
        else:
            tmp = pickle.load(open(pkl, "rb"))
            all_preds_labs["preds"] = np.concatenate(
                [all_preds_labs["preds"], tmp["preds"]], axis=0
            )
            all_preds_labs["labels"] = np.concatenate(
                [all_preds_labs["labels"], tmp["labels"]], axis=0
            )
            all_preds_labs["files"] = all_preds_labs["files"] + tmp["files"]
            all_preds_labs["datetimes"] = all_preds_labs["datetimes"] + tmp["datetimes"]

    utils.mkdir_p(outdir)
    pickle.dump(all_preds_labs, open(f"{outdir}/combined_preds_labs.pkl", "wb"))
    print(f"Wrote {outdir}/combined_preds_labs.pkl")


#################################################################################################
def combine_ltg_dbases(list_of_dbase_files, outdir):
    """
    Open and combine ltg_dbase.pkl files for a bunch of runs.
    Assumes that each dict has identical keys, and you should
    ensure that you can safely concatenate datetimes (i.e.,
    there should be no big gaps in time series. Disregard if
    you only want to aggregate lead-times and not score the data).
    """

    for i, pkl in enumerate(list_of_dbase_files):
        print(pkl)
        if i == 0:
            combined_dbase = pickle.load(open(pkl, "rb"))
        else:
            dd = pickle.load(open(pkl, "rb"))
            for key in dd:
                combined_dbase[key] += dd[key]

    utils.mkdir_p(outdir)
    pickle.dump(combined_dbase, open(f"{outdir}/combined_ltg_dbase.pkl", "wb"))
    print(f"Wrote {outdir}/combined_ltg_dbase.pkl")


#################################################################################################
def combine_ltg_dicts(list_of_dicts, outdir):
    """
    This is typically used to combine multiple ltg_score_dicts, like
    if you ran one per day for a month. We do this to get combined
    stats and lead-time.
    """

    for i, pkl in enumerate(list_of_dicts):
        print(pkl)
        if i == 0:
            combined_scores = pickle.load(open(pkl, "rb"))
        else:
            dd = pickle.load(open(pkl, "rb"))
            combined_scores["hits"] += dd["hits"]
            combined_scores["misses"] += dd["misses"]
            combined_scores["FAs"] += dd["FAs"]
            combined_scores["CNs"] += dd["CNs"]
            if dd["start_time"] < combined_scores["start_time"]:
                combined_scores["start_time"] = dd["start_time"]
            if dd["end_time"] > combined_scores["end_time"]:
                combined_scores["end_time"] = dd["end_time"]
            for th in combined_scores["leadtimes"]:
                combined_scores["leadtimes"][th] += dd["leadtimes"][th]

    for k, th in enumerate(combined_scores["thresholds"]):
        combined_scores["POD"][k] = combined_scores["hits"][k] / (
            combined_scores["hits"][k] + combined_scores["misses"][k]
        )
        combined_scores["FAR"][k] = combined_scores["FAs"][k] / (
            combined_scores["hits"][k] + combined_scores["FAs"][k]
        )
        combined_scores["CSI"][k] = combined_scores["hits"][k] / (
            combined_scores["hits"][k]
            + combined_scores["misses"][k]
            + combined_scores["FAs"][k]
        )
        combined_scores["med_lts"][k] = np.median(combined_scores["leadtimes"][th])

    utils.mkdir_p(outdir)
    pickle.dump(
        combined_scores, open(os.path.join(outdir, "combined_score_dict.pkl"), "wb")
    )
    print("Wrote " + os.path.join(outdir, "combined_score_dict.pkl"))


#################################################################################################
def write_ltg_stats(ltg_score_dict, ltg_dbase, outdir):

    pickle.dump(ltg_score_dict, open(os.path.join(outdir, "ltg_score_dict.pkl"), "wb"))
    pickle.dump(ltg_dbase, open(os.path.join(outdir, "ltg_dbase.pkl"), "wb"))

    meta = open(os.path.join(outdir, "stat_summary.txt"), "w")
    meta.write(
        "thresholds: "
        + (" ").join([str(thresh) for thresh in ltg_score_dict["thresholds"]])
        + "\n"
    )
    meta.write(
        "Correct nulls: "
        + (" ").join([str(int(cn)) for cn in ltg_score_dict["CNs"]])
        + "\n"
    )
    meta.write(
        "False alarms: "
        + (" ").join([str(int(fa)) for fa in ltg_score_dict["FAs"]])
        + "\n"
    )
    meta.write(
        "Hits: " + (" ").join([str(int(hit)) for hit in ltg_score_dict["hits"]]) + "\n"
    )
    meta.write(
        "Misses: "
        + (" ").join([str(int(miss)) for miss in ltg_score_dict["misses"]])
        + "\n"
    )
    meta.write(
        "POD: " + (" ").join(["{:.4f}".format(p) for p in ltg_score_dict["POD"]]) + "\n"
    )
    meta.write(
        "FAR: " + (" ").join(["{:.4f}".format(f) for f in ltg_score_dict["FAR"]]) + "\n"
    )
    meta.write(
        "CSI: " + (" ").join(["{:.4f}".format(c) for c in ltg_score_dict["CSI"]]) + "\n"
    )
    meta.write(
        "Median lead-time: "
        + (" ").join(["{:.1f}".format(ml) for ml in ltg_score_dict["med_lts"]])
        + "\n"
    )
    meta.close()


#################################################################################################
def first_flash_compare(topdir1, topdir2):
    """
    topdir1 and topdir2 are glob patterns.
    For each month for each model, plot the max CSI, 1st-flash median LT (at max CSI), and max-CSI prob. threshold.
    E.g., topdir1='/ships19/grain/probsevere/LC/tests/eval2021/v1/???21/' #jan21, feb21, etc.
    """

    ltg_score_dicts1 = glob.glob(f"{topdir1}/ltg_score_dict.pkl")
    ltg_score_dicts2 = glob.glob(f"{topdir2}/ltg_score_dict.pkl")
    if len(ltg_score_dicts1) == 0:
        print(f"No ltg_score_dict.pkls in {topdir1}")
        return
    if len(ltg_score_dicts2) == 0:
        print(f"No ltg_score_dict.pkls in {topdir2}")
        return

    months = [
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ]
    max_CSI1 = np.zeros(len(months))
    best_thresh1 = np.zeros(len(months))
    median_lt1 = np.zeros(len(months))
    for ii, mm in enumerate(months):
        # get pkl and index from ltg_score_dicts1
        for pkl_ind, pkl in enumerate(ltg_score_dicts1):
            if mm in os.path.basename(
                os.path.dirname(pkl)
            ):  # assumes month dir is just above ltg_score_dict.pkl
                scores = pickle.load(open(pkl, "rb"))
                best_ind = np.argmax(scores["CSI"])
                max_CSI1[ii] = scores["CSI"][best_ind]
                best_thresh1[ii] = scores["thresholds"][best_ind]
                median_lt1[ii] = scores["med_lts"][best_ind]
                break

    max_CSI2 = np.zeros(len(months))
    best_thresh2 = np.zeros(len(months))
    median_lt2 = np.zeros(len(months))
    for ii, mm in enumerate(months):
        # get pkl and index from ltg_score_dicts1
        for pkl_ind, pkl in enumerate(ltg_score_dicts2):
            if mm in os.path.basename(
                os.path.dirname(pkl)
            ):  # assumes month dir is just above ltg_score_dict.pkl
                scores = pickle.load(open(pkl, "rb"))
                best_ind = np.argmax(scores["CSI"])
                max_CSI2[ii] = scores["CSI"][best_ind]
                best_thresh2[ii] = scores["thresholds"][best_ind]
                median_lt2[ii] = scores["med_lts"][best_ind]
                break

    # make figure
    fig, ax1 = plt.subplots(nrows=1, ncols=1, figsize=(7, 5))

    months_upper = [mm.upper() for mm in months]
    ax1.set_xticks(np.arange(len(months)), labels=months_upper, fontsize=12)
    plt.ylim(0, 0.6)

    color1 = "blue"
    (csi1,) = plt.plot(
        np.arange(len(months)), max_CSI1, color=color1, lw=3, linestyle="-"
    )
    (bt1,) = plt.plot(
        np.arange(len(months)), best_thresh1, color=color1, lw=1, linestyle="-."
    )
    color2 = "red"
    (csi2,) = plt.plot(
        np.arange(len(months)), max_CSI2, color=color2, lw=3, linestyle="-"
    )
    (bt2,) = plt.plot(
        np.arange(len(months)), best_thresh2, color=color2, lw=1, linestyle="-."
    )

    ax2 = ax1.twinx()
    ax2.set_ylim([0, 50])
    ml1 = ax2.bar(np.arange(len(months)) - 0.15, median_lt1, width=0.3, color="#C8C8FF")
    ml2 = ax2.bar(np.arange(len(months)) + 0.15, median_lt2, width=0.3, color="#FFC8C8")
    ax2.set_ylabel("Median lead time @ best CSI thresh. [min] (bars)", fontsize=12)
    # bar handles
    ml1bar = plt.Rectangle((0, 0), 1, 1, color="#C8C8FF")
    ml2bar = plt.Rectangle((0, 0), 1, 1, color="#FFC8C8")

    # gridlines
    ax1.grid(axis="both", lw=0.5, linestyle=":", color="#9C9C9C")

    plt.title("First-flash metrics", fontsize=16)
    ax1.set_ylabel("Max CSI and best threshold []", fontsize=14)

    # legend
    leg = ax1.legend(
        [csi1, bt1, ml1bar, csi2, bt2, ml2bar],
        [
            "max CSI (v1)",
            "best prob. thresh (v1)",
            "median lead time (v1)",
            "max CSI (v2)",
            "best prob. thresh (v2)",
            "median lead time (v2)",
        ],
        prop={"size": 10},
        ncol=2,
        loc="upper center",
    )

    ax1.set_zorder(ax2.get_zorder() + 1)
    ax1.set_frame_on(False)
    plt.savefig("first_flash_graph.png", dpi=300, bbox_inches="tight")


#################################################################################################


def ltg_stats_figs(ltg_score_dict, outdir):

    import matplotlib.pyplot as plt

    # lead-time violin plot

    def adjacent_values(vals, q1, q3):
        upper_adjacent_value = q3 + (q3 - q1) * 1.5
        upper_adjacent_value = np.clip(upper_adjacent_value, q3, vals[-1])

        lower_adjacent_value = q1 - (q3 - q1) * 1.5
        lower_adjacent_value = np.clip(lower_adjacent_value, vals[0], q1)
        return lower_adjacent_value, upper_adjacent_value

    def set_axis_style(ax, labels):
        ax.get_xaxis().set_tick_params(direction="out")
        ax.xaxis.set_ticks_position("bottom")
        ax.set_xticks(np.arange(1, len(labels) + 1))
        ax.set_xticklabels(labels)
        ax.set_xlim(0.25, len(labels) + 0.75)

    fig, ax1 = plt.subplots(nrows=1, ncols=1, figsize=(7, 5))

    thresholds = ltg_score_dict["thresholds"][1:]  # skip 0.05 thresh
    leadtimes = ltg_score_dict["leadtimes"]
    data = [sorted(leadtimes[key]) for key in leadtimes][
        1:
    ]  # skip 0.05 thresh #data is a list of lists
    for ii in range(len(data)):
        data[ii] = [value for value in data[ii] if value <= 60]  # cap at 60

    # bw_method = 0.25 gives smoother distribution for quantized data (like LT for 5-min CONUS scans)
    # Just remove bw_method if you want to use the default "scott"
    parts = ax1.violinplot(
        data, showmeans=False, showmedians=False, showextrema=False, bw_method=0.25
    )

    for pc in parts["bodies"]:
        pc.set_facecolor("#D43F3A")
        pc.set_edgecolor("black")
        pc.set_alpha(1)

    quartile1, medians, quartile3 = [np.zeros(len(thresholds)) for _ in range(3)]
    for k in range(len(thresholds)):
        quartile1[k] = np.percentile(data[k], 25)
        medians[k] = np.percentile(data[k], 50)
        quartile3[k] = np.percentile(data[k], 75)
    # +/- 1.5*IQR
    # whiskers = np.array([adjacent_values(sorted_array, q1, q3) for sorted_array, q1, q3 in zip(data, quartile1, quartile3)])
    # whiskers_min, whiskers_max = whiskers[:, 0], whiskers[:, 1]
    # just hard-code min and max
    whiskers_min = np.array([2.5] * len(thresholds))
    whiskers_max = np.array([60] * len(thresholds))

    inds = np.arange(1, len(medians) + 1)
    ax1.scatter(inds, medians, marker="o", color="yellow", s=30, zorder=3)
    ax1.vlines(inds, quartile1, quartile3, color="k", linestyle="-", lw=5)
    # ax1.vlines(inds, whiskers_min, whiskers_max, color='k', linestyle='-', lw=1)

    ax1.set_xlabel("Probability threshold []")
    ax1.set_ylabel("Lead-time [min]")
    ax1.set_title("LightningCast lead-time to new lightning flash")

    plt.grid(axis="y", linestyle=":")

    # set style for the axes
    labels = [str(tt) for tt in thresholds]
    set_axis_style(ax1, labels)

    plt.savefig(
        os.path.join(outdir, "leadtime_violin.png"), dpi=300, bbox_inches="tight"
    )


###########################################################################################################
def seaborn_violin(score_dict1, score_dict2, outdir):

    import matplotlib.pyplot as plt
    import seaborn as sns
    import pandas as pd

    all_data = []

    # score_dict1 is with no onset criteria
    lts = score_dict1["leadtimes"]
    thresholds = score_dict1["thresholds"][1:]  # skip 0.05 thresh
    onset_criteria = "no"
    for tt in thresholds:
        for lt in lts[tt]:
            if lt <= 60:
                all_data.append([lt, tt, onset_criteria])

    # score_dict2 is with onset criteria
    onset_criteria = "yes"
    lts = score_dict2["leadtimes"]
    for tt in thresholds:
        for lt in lts[tt]:
            if lt <= 60:
                all_data.append([lt, tt, onset_criteria])

    # make dataframe
    df = pd.DataFrame(
        all_data,
        columns=["Lead-time [min]", "Probability threshold []", "Onset criterion"],
    )

    sns.set_theme(style="whitegrid")

    fig, ax1 = plt.subplots(nrows=1, ncols=1, figsize=(7, 5))

    ax = sns.violinplot(
        x="Probability threshold []",
        y="Lead-time [min]",
        hue="Onset criterion",
        data=df,
        palette="muted",
        bw=0.25,
        split=True,
        cut=0,
        inner="quartiles",
    )

    # change linestyles for quartiles
    for l in ax.lines[0::6]:
        l.set_linestyle("-")
        l.set_linewidth(1)
        l.set_color("#0000b5")  #'#4d4d4d')
    for l in ax.lines[2::6]:
        l.set_linestyle("-")
        l.set_linewidth(1)
        l.set_color("#0000b5")
    for l in ax.lines[1::6]:
        l.set_linestyle("-")
        l.set_linewidth(3)
        l.set_color("#0000b5")
    for l in ax.lines[3::6]:
        l.set_linestyle("-")
        l.set_linewidth(1)
        l.set_color("#b54200")
    for l in ax.lines[5::6]:
        l.set_linestyle("-")
        l.set_linewidth(1)
        l.set_color("#b54200")
    for l in ax.lines[4::6]:
        l.set_linestyle("-")
        l.set_linewidth(3)
        l.set_color("#b54200")
    #  for l in ax.lines[1::3]:
    #    l.set_linestyle('-')
    #    l.set_linewidth(3)
    #    l.set_color('black')

    ax1.set_title("LightningCast lead-time to new lightning flash", fontsize=12)

    plt.savefig(
        os.path.join(outdir, "sns_leadtime_violin.png"), dpi=300, bbox_inches="tight"
    )


###########################################################################################################
def get_ltg_marker():
    # see https://stackoverflow.com/questions/44726675/custom-markers-using-python-matplotlib
    u = np.array(
        [
            [-1.75, 5],
            [2.8, 5],
            [0.75, 1.5],
            [2.9, 1.5],
            [-3.75, -5.75],
            [-2, 0],
            [-4, 0],
            [-1.75, 5],
        ]
    )

    codes = [1] + [2] * (len(u) - 2) + [2]

    return mpath.Path(3 * u, codes, closed=False)  # False)


###########################################################################################################
def ts_setup(lines, timeseries, ll_bbox):
    """
    Simple initial setup.
    lines is a list of the lines read in from a static file or from the listened file
    timeseries is a list with the lat lon point and optionally a flag for "every time" or not
    ll_bbox is a list with the bounding box lat/lon coords. If not supplied, this will make one.
    """
    if not (
        ll_bbox
    ):  # if we want a timeseries, make a bounding box around the point (+/- 2deg)
        ll_bbox = [0, 0, 0, 0]
        ll_bbox[0] = timeseries[1] - 2
        ll_bbox[1] = timeseries[1] + 2  # lons
        ll_bbox[2] = timeseries[0] - 2
        ll_bbox[3] = timeseries[0] + 2  # lats
    ts_pltg, ts_pltg_radius, ltg_obs, ltg_obs_radius = [[] for _ in range(4)]
    # get all the datetimes upfront
    ts_dts = [datetime.strptime(ll, "%Y%m%d-%H%M%S") for ll in lines]
    # 3rd element is a flag for every time or not
    ts_every = True if (len(timeseries) == 3) else False
    crop = True

    return crop, ts_every, ts_dts, ts_pltg, ts_pltg_radius, ltg_obs, ltg_obs_radius


###########################################################################################################
def update_ltg_ts(
    new_preds,
    tsX,
    tsY,
    ts_pltg,
    ts_pltg_radius,
    ts_dts,
    line_cter,
    timeseries,
    outdir,
    ltg_grid,
    ltg_obs,
    ltg_obs_radius,
    everytime=False,
):
    """
    This function will update P(ltg) and FED time series and create a figure on the final (or every) interation.
    Params:
      new_preds: numpy array with the P(ltg) predictions
      tsX: The X coord of our point in image space.
      tsY: The Y coord of our point in image space.
      ts_pltg: A list containing the P(ltg) values in the time series (gets updated and passed back to calling function).
      ts_pltg_radius: A list containing the P(ltg) values within a radius of our point (gets updated and passed back).
      ts_dts: A list containing the datetimes in the time series.
      line_cter: The "line" or datetime that we're on.
      timeseries: a list with the [Lat,Lon] of our point.
      outdir: output directory
      ltg_grid: Usually 2D FED numpy array. Should already be cropped to match new_preds.
      ltg_obs: A list containing the binary marker for ltg present at the point of interest.
      ltg_obs_radius: A list containing the binary marker for ltg present within a radius.
      everytime: a boolean for if you want an image at every timestep.
    """
    ny, nx = new_preds.shape
    assert new_preds.shape == ltg_grid.shape

    r_pix = 13  # radius in pixels. Hardcoded for now.
    r_lab = "25"

    #  icp=[0,0,0,0,0,0,0,0,2,5,6,15,25,51,65,25,14,20,12]

    r_pix1 = 3
    ts_pltg.append(
        np.max(new_preds[tsY - r_pix1 : tsY + r_pix1, tsX - r_pix1 : tsX + r_pix1])
    )
    ts_pltg_radius.append(
        np.max(new_preds[tsY - r_pix : tsY + r_pix, tsX - r_pix : tsX + r_pix])
    )
    # add binary ltg marker (plotted at 90%) if ltg observation present
    if (
        np.max(ltg_grid[tsY - r_pix1 : tsY + r_pix1, tsX - r_pix1 : tsX + r_pix1])
        >= 0.1
    ):
        ltg_obs.append(90)
    else:
        ltg_obs.append(-10)
    if np.max(ltg_grid[tsY - r_pix : tsY + r_pix, tsX - r_pix : tsX + r_pix]) >= 0.1:
        ltg_obs_radius.append(90)
    else:
        ltg_obs_radius.append(-10)

    ndts = len(ts_dts)
    if line_cter == ndts or everytime:

        ts_pltg = [ppp * 100 for ppp in ts_pltg]
        ts_pltg_radius = [ppp * 100 for ppp in ts_pltg_radius]

        # make the image
        fig = plt.figure(
            num=None, figsize=(12, 7), dpi=300, facecolor="w", edgecolor="k"
        )  # , constrained_layout=True)
        gs = gridspec.GridSpec(ncols=1, nrows=1, figure=fig)
        ax_fs = 14
        ax = plt.subplot(gs[0:1, :])

        ylim = 100
        ax.set_ylim(0, ylim)
        ax.set_xlim(0, ndts - 1)  # ndts is number of datetimes
        ax.set_xticks(range(ndts))
        xticklabels = [ddd.strftime("%H:%M") for ddd in ts_dts]
        # should we plot every xlabel or every X?
        if ndts > 110:
            xticklabels = [""] + xticklabels[
                ::4
            ]  # plot every 4th time on the major ticks; not sure why the [''] is necessary, but it is.
            loc = MultipleLocator(4)
            ax.xaxis.set_major_locator(loc)
            majorloc = True
            # gridtype="major"
            minloc = MultipleLocator(1)
            ax.xaxis.set_minor_locator(minloc)
        elif ndts > 60:
            xticklabels = [""] + xticklabels[
                ::2
            ]  # plot every 2nd time on the major ticks; not sure why the [''] is necessary, but it is.
            loc = MultipleLocator(2)
            ax.xaxis.set_major_locator(loc)
            majorloc = True
            # gridtype="major"
            minloc = MultipleLocator(1)
            ax.xaxis.set_minor_locator(minloc)

        # xticklabels[0] = ndts[0].strftime('%Y-%m-%d %H:%M')
        ax.set_xticklabels(xticklabels, fontsize=ax_fs, rotation=90)
        ax.set_xlabel("Time (UTC)", fontsize=ax_fs)
        ax.set_yticks(range(0, 101, 10))
        ax.yaxis.grid(True, linestyle=":")
        ax.xaxis.grid(which="both", linestyle=":")
        ax.set_ylabel("Probability of lightning in next 60 min [%]", fontsize=ax_fs)

        alpha = 0.15
        (line1,) = ax.plot(range(ndts), [10] * ndts, color="blue", lw=1, linestyle="-")
        (line2,) = ax.plot(range(ndts), [25] * ndts, color="cyan", lw=1, linestyle="-")
        plt.fill_between(
            range(ndts), y1=[10] * ndts, y2=[25] * ndts, facecolor="blue", alpha=alpha
        )
        (line3,) = ax.plot(
            range(ndts), [50] * ndts, color="#00FF00", lw=1, linestyle="-"
        )
        plt.fill_between(
            range(ndts), y1=[25] * ndts, y2=[50] * ndts, facecolor="cyan", alpha=alpha
        )
        (line4,) = ax.plot(
            range(ndts), [75] * ndts, color="#FF00FF", lw=1, linestyle="-"
        )
        plt.fill_between(
            range(ndts),
            y1=[50] * ndts,
            y2=[75] * ndts,
            facecolor="#00FF00",
            alpha=alpha,
        )
        (line5,) = ax.plot(
            range(ndts), [100] * ndts, color="#FF00FF", lw=1, linestyle="-"
        )
        plt.fill_between(
            range(ndts),
            y1=[75] * ndts,
            y2=[100] * ndts,
            facecolor="#FF00FF",
            alpha=alpha,
        )

        if len(ts_pltg) < ndts:  # indicates we're making an image at every time
            # pltgLine, = ax.plot(range(len(ts_pltg)),ts_pltg,color='red',lw=5.0,linestyle='-')
            (pltgRadiusLine,) = ax.plot(
                range(len(ts_pltg)), ts_pltg_radius, color="red", lw=5.0, linestyle="-"
            )
            # icpline, = ax.plot(range(len(ts_pltg)),icp[0:len(ts_pltg)],color='blue',lw=5,linestyle='-')
            vertical_line = ax.plot(
                [len(ts_pltg) - 1, len(ts_pltg) - 1],
                [0, ylim],
                color="black",
                lw=2,
                linestyle="-",
            )
        else:
            # pltgLine, = ax.plot(range(ndts),ts_pltg,color='red',lw=5.0,linestyle='-')
            # icpline, = ax.plot(range(ndts),icp,color='blue',lw=5,linestyle='-')
            (pltgRadiusLine,) = ax.plot(
                range(ndts), ts_pltg_radius, color="red", lw=5.0, linestyle="-"
            )

        # plog the ltg observations. The -0.5 is because GLM FED is a 5-min aggregation from previous time step.
        ltg_marker = get_ltg_marker()
        fc = "orange"
        ec = "orange"
        if len(ts_pltg) < ndts:  # indicates we're making an image at every time
            (ltgObs,) = ax.plot(
                np.arange(-0.5, len(ltg_obs) - 0.5, 1),
                ltg_obs,
                linestyle="None",
                color=fc,
                marker=ltg_marker,
                markersize=30,
            )
            (ltgObsRadius,) = ax.plot(
                np.arange(-0.5, len(ltg_obs) - 0.5, 1),
                ltg_obs_radius,
                linestyle="None",
                color=fc,
                marker=ltg_marker,
                markersize=15,
            )
        else:
            (ltgObs,) = ax.plot(
                np.arange(-0.5, ndts - 0.5, 1),
                ltg_obs,
                linestyle="None",
                color=fc,
                marker=ltg_marker,
                markersize=30,
            )
            (ltgObsRadius,) = ax.plot(
                np.arange(-0.5, ndts - 0.5, 1),
                ltg_obs_radius,
                linestyle="None",
                color=fc,
                marker=ltg_marker,
                markersize=15,
            )

        # leg = ax.legend([icpline,pltgLine,pltgRadiusLine,ltgObs,ltgObsRadius],
        #  ['ICP within 6 km','P(LTG) within 6 km','P(LTG) within '+r_lab+' km',
        #   'GLM flash obs. within 6 km','GLM flash obs. within '+r_lab+' km'],prop={'size':12},ncol=1,
        #   loc='lower right') #,bbox_to_anchor=(0.75, 0.1))
        # leg = ax.legend([pltgLine,pltgRadiusLine,ltgObs,ltgObsRadius],
        leg = ax.legend(
            [pltgRadiusLine, ltgObs, ltgObsRadius],
            [  #'P(LTG) within 6 km',
                "P(LTG) within " + r_lab + " km",
                "GLM flash obs. within 6 km",
                "GLM flash obs. within " + r_lab + " km",
            ],
            prop={"size": 12},
            ncol=1,
            loc="lower right",
        )  # ,bbox_to_anchor=(0.75, 0.1))
        leg.set_zorder(50)

        ax.text(0.01, 0.02, "LOW", transform=ax.transAxes, fontsize=12, rotation=90)
        ax.text(
            0.01, 0.12, "ELEVATED", transform=ax.transAxes, fontsize=10, rotation=90
        )
        ax.text(0.01, 0.3, "ENHANCED", transform=ax.transAxes, fontsize=12, rotation=90)
        ax.text(
            0.01, 0.55, "MODERATE", transform=ax.transAxes, fontsize=12, rotation=90
        )
        ax.text(0.01, 0.83, "HIGH", transform=ax.transAxes, fontsize=14, rotation=90)

        lattype = "N" if (timeseries[0] > 0) else "S"
        lontype = "W" if (timeseries[1] < 0) else "E"
        plt.suptitle("ProbSevere LightningCast time series", fontsize=16, y=0.96)
        plt.title(
            "Point of interest: "
            + "{:.2f}".format(timeseries[0])
            + lattype
            + " "
            + "{:.2f}".format(timeseries[1])
            + lontype
            + "; Initial date: "
            + ts_dts[0].strftime("%Y-%m-%d"),
            fontsize=ax_fs,
        )
        lat = "{:.2f}".format(timeseries[0])
        lon = "{:.2f}".format(timeseries[1])
        if everytime:
            time1 = ts_dts[len(ts_pltg) - 1].strftime("%Y%m%d-%H%M")
            plt.savefig(
                os.path.join(outdir, ("_").join(["PLTG", time1, lat, lon]) + ".png"),
                format="png",
                bbox_inches="tight",
            )
        else:
            time1 = ts_dts[0].strftime("%Y%m%d-%H%M")
            time2 = ts_dts[-1].strftime("%Y%m%d-%H%M")
            plt.savefig(
                os.path.join(
                    outdir, ("_").join(["PLTG", time1, time2, lat, lon]) + ".png"
                ),
                format="png",
                bbox_inches="tight",
            )
