import time
import numpy as np
import xarray as xr
import pandas as pd
import os, sys
import pickle
from datetime import datetime, timedelta
import glob
import pyresample as pr

sys.path.insert(0, "..")
import utils
from netCDF4 import Dataset
import shutil
import argparse
from subprocess import Popen, call, PIPE

PWD = os.environ["PWD"] + "/"
import traceback

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"  # visible in this process + all children

# CONSTANTS
# c = 2.99792458e8 #speed of light (m/s)
# h = 6.62607015e-34 #Planck's constant (J s)
# k = 1.38064852e-23 #Boltzmann's constant (J/K)
# l14 = 11.2e-6 #central wavelength of CH14 (m)

# formula for converting from radiance to BT.
# see https://www.star.nesdis.noaa.gov/goesr/docs/ATBD/Imagery.pdf
# T = [ fk2 / (alog((fk1 / Lλ) + 1)) - bc1 ] / bc2   eqn (3-5)

# for CH02 radiance to reflectance:
# see pdf above pp. 20-22
###############################################################################################################
def assign_data(netcdf, data, ch, startY, endY, startX, endX, PWD, ii):

    baddata = False

    nc = Dataset(PWD + os.path.basename(netcdf), "r")
    if ch == "02":
        radiance = nc.variables["Rad"][startY * 4 : endY * 4, startX * 4 : endX * 4]
    elif ch == "03" or ch == "01" or ch == "05":
        radiance = nc.variables["Rad"][startY * 2 : endY * 2, startX * 2 : endX * 2]
    else:
        radiance = nc.variables["Rad"][startY:endY, startX:endX]

    if int(ch) <= 6:
        # convert vis radiance to reflectance
        refl = radiance * nc.variables["kappa0"][:]
        data["CH" + ch][ii] = refl
        if np.max(np.abs(refl)) > 2:
            baddata = True
    else:
        # convert IR radiance to brightness temperature
        bt = (
            nc.variables["planck_fk2"][:]
            / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
            - nc.variables["planck_bc1"][:]
        ) / nc.variables["planck_bc2"][:]
        data["CH" + ch][ii] = bt
        if np.max(np.abs(bt)) > 500:
            baddata = True

    # close
    nc.close()

    return baddata


###############################################################################################################
def read_netcdf(file, data={}, global_atts={}):

    if file[-3:] == ".gz":
        gzip = 1
        unzipped_file = file[0 : len(file) - 3]
        the_file = unzipped_file
        try:
            s = call(["gunzip", "-f", file])
        except:
            print("cannot open " + file)
            return False
    else:
        gzip = 0
        the_file = file

    try:
        openFile = Dataset(the_file, "r")
    except (OSError, IOError):
        print(traceback.format_exc())
        return False

    for v in data:
        if v in openFile.variables:
            data[v] = openFile.variables[v][:]

    # get global attributes
    for gatt in global_atts:
        if gatt in openFile.ncattrs():
            global_atts[gatt] = getattr(
                openFile, gatt
            )  # global_atts is modified and returned
        else:
            print(gatt + " not in " + file)
            openFile.close()
            return False

    openFile.close()

    # --gzip file after closing?
    if gzip:
        s = Popen(
            ["gzip", "-f", unzipped_file]
        )  # ==will run in background since it doesn't need output from command.
        # s = call(['gzip','-f',unzipped_file])


###############################################################################################################
def get_glm_data(glm_file, data={}, global_atts={}, plax=None):

    if glm_file[-3:] == ".gz":
        gzip = 1
        unzipped_file = glm_file[0 : len(glm_file) - 3]
        the_file = unzipped_file
        try:
            s = call(["gunzip", "-f", glm_file])
        except:
            print("cannot open " + glm_file)
            return False
    else:
        gzip = 0
        the_file = glm_file

    try:
        openFile = Dataset(the_file, "r")
    except (OSError, IOError):
        print(traceback.format_exc())
        return False

    glmvars = [
        "flash_extent_density",
        "flash_centroid_density",
        "total_energy",
        "average_flash_area",
    ]
    for v in glmvars:
        if v in openFile.variables:
            if plax:  # this only works for 2D vars
                print("Parallax correcting " + v)
                tmpvar = openFile.variables[v][:]
                plax_var = np.zeros(np.shape(tmpvar))
                plax_var[plax] = np.ravel(tmpvar)
                data[v] = plax_var
            else:
                data[v] = openFile.variables[v][:]

    # get global attributes
    # for gatt in global_atts:
    #  if(gatt in openFile.ncattrs()):
    #    global_atts[gatt] = getattr(openFile,gatt) #global_atts is modified and returned
    #  else:
    #    print(gatt + ' not in ' + glm_file)
    #    openFile.close()
    #    return False

    openFile.close()

    # --gzip file after closing?
    if gzip:
        s = Popen(
            ["gzip", "-f", unzipped_file]
        )  # ==will run in background since it doesn't need output from command.
        # s = call(['gzip','-f',unzipped_file])

    return True


###############################################################################################################


def main():

    vars_to_write = {
        "CH02": {
            "atts": {
                "desc": "t by x by y slice of 0.65-micron refl from ABI",
                "units": "dimensionless",
            }
        },
        "CH05": {
            "atts": {
                "desc": "t by x by y slice of 1.6-micron refl from ABI",
                "units": "dimensionless",
            }
        },
        "CH06": {
            "atts": {
                "desc": "t by x by y slice of 2.26-micron refl from ABI",
                "units": "dimensionless",
            }
        },
        "CH08": {
            "atts": {
                "desc": "t by x by y slice of 6.2-micron BT from ABI",
                "units": "K",
            }
        },
        "CH10": {
            "atts": {
                "desc": "t by x by y slice of 7.3-micron BT from ABI",
                "units": "K",
            }
        },
        "CH11": {
            "atts": {
                "desc": "t by x by y slice of 8.6-micron BT from ABI",
                "units": "K",
            }
        },
        "CH13": {
            "atts": {
                "desc": "t by x by y slice of 10.3-micron BT from ABI",
                "units": "K",
            }
        },
        "CH15": {
            "atts": {
                "desc": "t by x by y slice of 12.4-micron BT from ABI",
                "units": "K",
            }
        },
        "CH16": {
            "atts": {
                "desc": "t by x by y slice of 12.4-micron BT from ABI",
                "units": "K",
            }
        },
        "FED": {
            "atts": {
                "desc": "GLM flash extent density, accumulated over 5 min",
                "units": "fl",
            }
        },
        "MAXFED_5to60": {
            "atts": {"desc": "max FED from 5 to 60 min in the future", "units": "fl"}
        },
        "MAXFED_5to45": {
            "atts": {"desc": "max FED from 5 to 45 min in the future", "units": "fl"}
        },
        "MAXFED_5to30": {
            "atts": {"desc": "max FED from 5 to 30 min in the future", "units": "fl"}
        },
        "ENI": {
            "atts": {
                "desc": "ENI flash density, accumulated over 2 min",
                "units": "fl/km^2/min",
            }
        },
        "MAXENI_0to60": {
            "atts": {
                "desc": "max ENI from 0 to 60 min in the future",
                "units": "fl/km^2/min",
            }
        },
        "MAXENI_0to45": {
            "atts": {
                "desc": "max ENI from 0 to 45 min in the future",
                "units": "fl/km^2/min",
            }
        },
        "MAXENI_0to30": {
            "atts": {
                "desc": "max ENI from 0 to 30 min in the future",
                "units": "fl/km^2/min",
            }
        },
        "mlcape": {
            "atts": {
                "desc": "mixed-layer CAPE (HRRR 2D surface levels hourly files)",
                "units": "J/kg",
            }
        },
        "u925": {
            "atts": {
                "desc": "U-comp of 925mb wind (HRRR 2D surface levels hourly files)",
                "units": "m/s",
            }
        },
        "v925": {
            "atts": {
                "desc": "V-comp of 925mb wind (HRRR 2D surface levels hourly files)",
                "units": "m/s",
            }
        },
        "u850": {
            "atts": {
                "desc": "U-comp of 850mb wind (HRRR 2D surface levels hourly files)",
                "units": "m/s",
            }
        },
        "v850": {
            "atts": {
                "desc": "V-comp of 850mb wind (HRRR 2D surface levels hourly files)",
                "units": "m/s",
            }
        },
        "u700": {
            "atts": {
                "desc": "U-comp of 700mb wind (HRRR 2D surface levels hourly files)",
                "units": "m/s",
            }
        },
        "v700": {
            "atts": {
                "desc": "V-comp of 700mb wind (HRRR 2D surface levels hourly files)",
                "units": "m/s",
            }
        },
        "u500": {
            "atts": {
                "desc": "U-comp of 500mb wind (HRRR 2D surface levels hourly files)",
                "units": "m/s",
            }
        },
        "v500": {
            "atts": {
                "desc": "V-comp of 500mb wind (HRRR 2D surface levels hourly files)",
                "units": "m/s",
            }
        },
        "GHRSST": {
            "atts": {
                "desc": "Global Hi-Res Sea-Surface Temperature (native res: 5km)",
                "units": "K",
            }
        },
        "dates": {"atts": {"desc": "UTC yyymmdd"}},
        "times": {"atts": {"desc": "UTC hhmm"}},
    }

    # just some placeholders for now
    global_atts = {"ABI_time": -1}

    satdatadir = "/arcdata/goes/grb/goes16/"
    copy = True
    # sstdir = '/iliad/grain/common/oconus_ps/OISST/' #check GHRSST?
    ghrsst_dir = "/iliad/grain/common/oconus_ps/GHRSST_5km/"
    glmdir = "/apollo/grain/saved_probsevere_data/lightning/"
    enidir = glmdir
    hrrrdir = "/data/jcintineo/HRRR/remapped/"
    outbase = "/apollo/grain/jcintineo/AI/OCONUS/seg/netcdf400x400_eni_hrrr_ghrsst/"

    abi_channels = ["02", "05", "06", "08", "10", "11", "13", "15", "16"]
    glm_channels = ["FED"]

    # get ABI lats/lons
    area_def, ny, nx, abi_lons, abi_lats = utils.get_area_definition(
        "../static/GOES_East.nc", return_latlons=True
    )
    # get land/sea mask
    mask_nc = Dataset("../static/mask_GOES-East_CONUS_25km.nc", "r")
    landsea_mask = mask_nc.variables["mask"][
        :
    ]  # this is on the ABI grid with effective resolution of 25km.
    mask_nc.close()

    predict_leadtime = 60  # in min
    pred_steps = predict_leadtime / 5 + 1
    global_atts["predict_leadtime"] = predict_leadtime

    ltg_frac_thresh = 0.005

    # with respect to GOES-East 2km ABI CONUS grid
    startX = 1510
    endX = 1910  # 2310
    startY = 450  # 50
    endY = 850  # this should be north of the "Bermuda bars/strobes"

    abi_lons = abi_lons[startY:endY, startX:endX]
    abi_lats = abi_lats[startY:endY, startX:endX]
    sat_grid = pr.geometry.GridDefinition(lons=abi_lons, lats=abi_lats)
    sat_proj_sst_grid = pr.geometry.GridDefinition(
        lons=abi_lons[::2, ::2], lats=abi_lats[::2, ::2]
    )  # reducing to 4km res

    # radar grid
    NWlat = 51
    NWlon = -125
    dy = dx = 0.01
    nlon = 5900
    nlat = 2700
    rad_lats = np.zeros((nlat, nlon))
    rad_lons = np.zeros((nlat, nlon))
    for ii in range(nlon):
        rad_lons[:, ii] = NWlon + ii * dx
    for ii in range(nlat):
        rad_lats[ii, :] = NWlat - ii * dy
    eni_grid = pr.geometry.GridDefinition(lons=rad_lons, lats=rad_lats)

    sat_steps = [20, 15, 10, 5]  # these are the delta Ts we're subtracting
    nframes = len(sat_steps) + 1  # add in t0, as well

    curr_hrrr_file = ""
    curr_sst_file = ""

    # instantiate blank arrays
    data = {}
    for ch in abi_channels:
        if ch == "02":
            data["CH02"] = np.zeros(
                (nframes, endY * 4 - startY * 4, endX * 4 - startX * 4)
            )
        elif ch == "05" or ch == "01" or ch == "03":
            data["CH" + ch] = np.zeros(
                (nframes, endY * 2 - startY * 2, endX * 2 - startX * 2)
            )
        else:
            data["CH" + ch] = np.zeros((nframes, endY - startY, endX - startX))
    for ch in glm_channels:
        data[ch] = np.zeros((nframes, endY - startY, endX - startX))

    # for d in [20180517,20180521,20180528,20180604,20180615,20180709,20180718,20180725,20180809,20180818,20180824]:
    # for d in [20180517,20180521,20180525,20180528,20180604,20180615,20180616,20180623,20180709,20180710,20180718,20180725,20180809,20180818,20180823,20180824]:
    for d in (
        list(range(20180506, 20180532))
        + list(range(20180601, 20180631))
        + list(range(20180701, 20180732))
        + list(range(20180801, 20180832))
    ):

        strdate = str(d)

        topdt = datetime.strptime(strdate, "%Y%m%d")

        satdir1 = topdt.strftime(satdatadir + "/%Y/%Y_%m_%d_%j/abi/L1b/RadC/")

        # sat files for the day #CHANGE THE 2* HERE ??????
        satfiles = glob.glob(
            os.path.join(satdir1, topdt.strftime("OR_ABI-L1b-RadC-M*C13_G16_s%Y%j*"))
        )
        satfiles = np.sort(satfiles)
        # satfiles = satfiles[0::2] #every other satfile
        satdts = np.array(
            [
                datetime.strptime(
                    os.path.basename(x).split("_s")[1][0:13], "%Y%j%H%M%S"
                )
                for x in satfiles
            ]
        )

        # for each satdt, get the t0, and t-sat_step grids
        reset = True
        for dt in satdts:
            print("reset:", reset)
            cont = False
            yyyymmdd = dt.strftime("%Y%m%d")
            outts = dt.strftime("%Y%m%d-%H%M%S")

            # get all of the data first. Then do extraction.

            # first check if we have enough lightning to make it worth our while
            # get GLM data
            # we can't necessarily assume sat files will be every 5 min
            future_glm = []
            tmpdt = dt
            print("getting glm data")
            while len(future_glm) < pred_steps:
                netcdfpattern = tmpdt.strftime(
                    os.path.join(
                        glmdir, "%Y%m%d", "GLM", "goes_east", "agg", "%Y%m%d-%H%M*"
                    )
                )
                try:
                    netcdf = glob.glob(netcdfpattern)[0]
                except IndexError as err:
                    print(traceback.format_exc())
                    print(netcdfpattern)
                    cont = True
                    break
                else:
                    glm_data = {}
                    new_glm_data = get_glm_data(netcdf, data=glm_data)
                    future_glm.append(
                        glm_data["flash_extent_density"][startY:endY, startX:endX]
                    )
                    tmpdt = tmpdt + timedelta(minutes=5)
            if cont:
                continue  # move on to next time, since we don't have all the truth data
            print("done")

            all_future_glm = np.stack(
                future_glm[1:]
            )  # don't stack 0th time since it's ltg from t0-5 min to t0
            max_future_glm = np.max(all_future_glm, axis=0)
            tot_pix = max_future_glm.shape[0] * max_future_glm.shape[1]
            good_ind = np.where(max_future_glm > 0)
            if len(good_ind[0]) > 0:
                ltg_frac = len(good_ind[0]) / tot_pix
            else:
                ltg_frac = 0

            if ltg_frac < ltg_frac_thresh:
                print(dt, "ltg_frac = ", np.round(ltg_frac, 6), "...not enough ltg")
                continue
            else:
                print(dt, "ltg_frac = ", np.round(ltg_frac, 6))
                max_glm_5to30 = np.max(all_future_glm[:6], axis=0)
                max_glm_5to45 = np.max(all_future_glm[:9], axis=0)

                # ltg frac is good, so get prev glm
                data["FED"][-1] = future_glm[0]
                for ii, mm in enumerate(sat_steps):
                    tmpdt = dt - timedelta(minutes=mm)
                    netcdfpattern = tmpdt.strftime(
                        os.path.join(
                            glmdir, "%Y%m%d", "GLM", "goes_east", "agg", "%Y%m%d-%H%M*"
                        )
                    )
                    try:
                        netcdf = glob.glob(netcdfpattern)[0]
                    except IndexError as err:
                        print(traceback.format_exc())
                        print(netcdfpattern)
                        cont = True
                        break
                    else:
                        glm_data = {}
                        new_glm_data = get_glm_data(netcdf, data=glm_data)
                        data["FED"][ii] = glm_data["flash_extent_density"][
                            startY:endY, startX:endX
                        ]

                if cont:
                    continue  # move on to next time, since we don't have all the previous glm data

                data["MAXFED_5to60"] = max_future_glm
                data["MAXFED_5to45"] = max_glm_5to45
                data["MAXFED_5to30"] = max_glm_5to30
                data["dates"] = np.arange(nframes)
                data["times"] = np.arange(nframes)
                data["dates"][-1] = int(dt.strftime("%Y%m%d"))
                data["times"][-1] = int(dt.strftime("%H%M"))
                for ii, mm in enumerate(sat_steps):
                    data["dates"][ii] = int(
                        (dt - timedelta(minutes=mm)).strftime("%Y%m%d")
                    )
                    data["times"][ii] = int(
                        (dt - timedelta(minutes=mm)).strftime("%H%M")
                    )

            # get ENI data
            t0 = time.time()
            print("getting ENI data")
            prev_eni = []
            for ee in sat_steps + [0]:  # for each previous step
                for tt in [0, 1, 2, 3]:  # offset in min
                    tmpdt = dt - timedelta(minutes=(ee + tt))
                    netcdfpattern = tmpdt.strftime(
                        os.path.join(
                            enidir,
                            "%Y%m%d",
                            "ENI_LightningDensity",
                            "002_min",
                            "%Y%m%d-%H%M*",
                        )
                    )
                    try:
                        netcdf = glob.glob(netcdfpattern)[0]
                    except IndexError as err:
                        if tt == 3:
                            print(traceback.format_exc())
                            print(netcdfpattern)
                            cont = True
                            break
                    else:
                        eni_nc = Dataset(netcdf)
                        new_eni_data = eni_nc.variables["ENI_LightningDensity"][:]
                        eni_nc.close()
                        prev_eni.append(
                            pr.kd_tree.resample_nearest(
                                eni_grid,
                                new_eni_data,
                                sat_grid,
                                radius_of_influence=4000,
                                fill_value=-999,
                            )
                        )
                        break

                if cont:
                    break  # can't find ENI data

            if cont:
                continue  # move on to next time, since we don't have all the truth data

            data["ENI"] = np.stack(prev_eni)  # includes 't0' as the -1 slice

            # get future ENI data
            start_dt = dt
            end_dt = start_dt + timedelta(minutes=60)
            valid_eni = []
            tmpdt = start_dt + timedelta(minutes=1)
            while start_dt < tmpdt <= end_dt:
                eni_patt = tmpdt.strftime(
                    os.path.join(
                        enidir,
                        "%Y%m%d",
                        "ENI_LightningDensity",
                        "002_min",
                        "%Y%m%d-%H%M*",
                    )
                )
                print(dt, eni_patt)
                try:
                    eni_file = glob.glob(eni_patt)[0]
                except IndexError:
                    pass
                else:
                    eni_nc = Dataset(eni_file)
                    valid_eni.append(eni_nc.variables["ENI_LightningDensity"][:])
                    eni_nc.close()
                tmpdt = tmpdt + timedelta(minutes=1)

            try:
                future_eni = np.stack(valid_eni)
            except ValueError as err:  # no values in valid_eni
                print(err)
                print("move on to next dt")
                continue

            maxeni_60min = np.max(future_eni, axis=0)
            maxeni_45min = np.max(future_eni[0:23], axis=0)
            maxeni_30min = np.max(future_eni[0:16], axis=0)

            # now remap the eni data
            remapped_eni_60min = pr.kd_tree.resample_nearest(
                eni_grid,
                maxeni_60min,
                sat_grid,
                radius_of_influence=4000,
                fill_value=-999,
            )
            remapped_eni_45min = pr.kd_tree.resample_nearest(
                eni_grid,
                maxeni_45min,
                sat_grid,
                radius_of_influence=4000,
                fill_value=-999,
            )
            remapped_eni_30min = pr.kd_tree.resample_nearest(
                eni_grid,
                maxeni_30min,
                sat_grid,
                radius_of_influence=4000,
                fill_value=-999,
            )
            data["MAXENI_0to60"] = remapped_eni_60min
            data["MAXENI_0to45"] = remapped_eni_45min
            data["MAXENI_0to30"] = remapped_eni_30min

            # get satellite data
            broke = False
            for ch in abi_channels:
                if (
                    reset
                ):  # firstdt in the day or found some missing/bad data, so just do each dt frame
                    for ii, tt in enumerate(sat_steps + [0]):
                        newdt = dt - timedelta(minutes=tt)

                        netcdfpattern = newdt.strftime(
                            satdatadir
                            + "/%Y/%Y_%m_%d_%j/abi/L1b/RadC/OR_ABI-L1b-RadC-M*C"
                            + ch
                            + "_G16_s%Y%j%H%M*"
                        )
                        try:
                            netcdf = glob.glob(netcdfpattern)[0]
                        except IndexError as err:
                            # a desired channel does not exist
                            print(traceback.format_exc())
                            print("Moving on to next sat DT...")
                            broke = True
                            break
                        # copy file from satdatadir
                        try:
                            shutil.copy(netcdf, PWD)
                        except (PermissionError, OSError, IOError) as err:
                            print(traceback.format_exc())
                            print("Moving on to next sat DT...")
                            broke = True
                            break
                        else:
                            baddata = assign_data(
                                netcdf, data, ch, startY, endY, startX, endX, PWD, ii
                            )
                            os.remove(PWD + os.path.basename(netcdf))
                            if baddata:
                                broke = True
                                break
                    if (
                        broke
                    ):  # either bad data or the channel doesn't exist or we can't copy the data
                        break  # for ch in abi_channels

                else:  # reset = False
                    # move data temporally...0 to -5, -5 to -10, etc.
                    # I think we don't do this for FED because delta T for sat could be more than 5 min.
                    # I think we could do this with FED if everything has delta T = 5 min.
                    data["CH" + ch][0:-1] = data["CH" + ch][1:]

                    newdt = dt
                    netcdfpattern = newdt.strftime(
                        satdatadir
                        + "/%Y/%Y_%m_%d_%j/abi/L1b/RadC/OR_ABI-L1b-RadC-M*C"
                        + ch
                        + "_G16_s%Y%j%H%M*"
                    )
                    try:
                        netcdf = glob.glob(netcdfpattern)[0]
                    except IndexError as err:
                        # a desired channel does not exist
                        print(traceback.format_exc())
                        print("Moving on to next sat DT...")
                        broke = True
                    # copy file from satdatadir
                    try:
                        shutil.copy(netcdf, PWD)
                    except (PermissionError, OSError, IOError) as err:
                        print(traceback.format_exc())
                        print("Moving on to next sat DT...")
                        broke = True
                    else:
                        baddata = assign_data(
                            netcdf, data, ch, startY, endY, startX, endX, PWD, -1
                        )  # -1 indicates last frame
                        os.remove(PWD + os.path.basename(netcdf))
                        if baddata:
                            broke = True
                    if (
                        broke
                    ):  # either bad data or the channel doesn't exist or we can't copy the data
                        break  # for ch in abi_channels

            if broke:
                continue  # continue to next satdt
            else:
                reset = False

            # get HRRR data
            hrrr_dt = dt - timedelta(hours=1)  # to simulate realtime
            hrrr_pattern = hrrr_dt.strftime(
                os.path.join(hrrrdir, "%Y%m%d", "HRRR_remapped_%Y%m%d_%H0000.nc*")
            )
            netcdf = glob.glob(hrrr_pattern)[0]
            if netcdf != curr_hrrr_file:
                curr_hrrr_file = netcdf  # reset
                print("opening " + netcdf)
                hrrr_data = {
                    "mlcape": -1,
                    "u925": -1,
                    "v925": -1,
                    "u850": -1,
                    "v850": -1,
                    "u700": -1,
                    "v700": -1,
                    "u500": -1,
                    "v500": -1,
                }
                hrrr_status = read_netcdf(netcdf, data=hrrr_data)
                for hv in hrrr_data:
                    data[hv] = hrrr_data[hv][startY:endY, startX:endX]

            # GHRSST
            sst_dt = dt - timedelta(days=1)  # to simulate realtime
            dkey = sst_dt.strftime("%Y%m%d")
            # find and open sst file
            try:
                ncfile = glob.glob(os.path.join(ghrsst_dir, dkey + "*.nc"))[0]
            except IndexError:
                sst_dt = datetime.strptime(str(dates[i]), "%Y%m%d") - timedelta(
                    days=1
                )  # try 2 days ago
                ncfile = glob.glob(
                    os.path.join(ghrsst_dir, sst_dt.strftime("%Y%m%d") + "*.nc")
                )[0]
            if netcdf != curr_sst_file:
                curr_sst_file = netcdf  # reset
                print("opening " + ncfile)
                sst_nc = Dataset(ncfile, "r")
                try:
                    sst_lat
                except NameError:
                    sst_lat = sst_nc.variables["lat"][:]
                    sst_lon = sst_nc.variables["lon"][:]
                    sst_lon, sst_lat = np.meshgrid(sst_lon, sst_lat)
                    sst_grid = pr.geometry.GridDefinition(lons=sst_lon, lats=sst_lat)
                sst_data = sst_nc.variables["analysed_sst"][0, :, :]
                sst_nc.close()
                sst_data[sst_data > 500] = 0
                sst_data[sst_data < 0] = 0  # fix land mask
                # only remap if a new file
                data["GHRSST"] = pr.kd_tree.resample_nearest(
                    sst_grid,
                    sst_data,
                    sat_proj_sst_grid,
                    radius_of_influence=4000,
                    fill_value=-999,
                )

            global_atts["ABI_time"] = dt.strftime("%Y%m%d-%H%M%S UTC")

            # write the final netCDFs for our database
            datasets = {}
            for key, val in data.items():
                if key != "dates" and key != "times":
                    datasets[key] = {"data": val.astype(np.float32)}
                else:
                    datasets[key] = {"data": val}

            dims = {}
            dims["t"], dims["x"], dims["y"] = np.shape(datasets["CH13"]["data"])
            _, dims["x05"], dims["y05"] = np.shape(datasets["CH02"]["data"])
            _, dims["x10"], dims["y10"] = np.shape(datasets["CH05"]["data"])
            dims["sst_x"], dims["sst_y"] = np.shape(datasets["GHRSST"]["data"])

            output_file = os.path.join(
                outbase, yyyymmdd, ("_").join(["goes16", "conus", outts + ".netcdf"])
            )

            for v in datasets:
                datasets[v]["atts"] = vars_to_write[v][
                    "atts"
                ]  # this sets the metadata for v
                if v == "CH02":
                    currdim = ("t", "x05", "y05")
                elif v in ["CH01", "CH03", "CH05"]:
                    currdim = ("t", "x10", "y10")
                elif v in [
                    "CH04",
                    "CH06",
                    "CH07",
                    "CH08",
                    "CH09",
                    "CH10",
                    "CH11",
                    "CH12",
                    "CH13",
                    "CH14",
                    "CH15",
                    "CH16",
                    "FED",
                    "FCD",
                    "TOE",
                    "AFA",
                    "ENI",
                ]:
                    currdim = ("t", "x", "y")
                elif v.startswith("MAXFED"):
                    currdim = ("x", "y")
                elif v.startswith("MAXENI"):
                    currdim = ("x", "y")
                elif v in [
                    "mlcape",
                    "u925",
                    "v925",
                    "u850",
                    "v850",
                    "u700",
                    "v700",
                    "u500",
                    "v500",
                ]:
                    currdim = ("x", "y")
                elif v == "GHRSST":
                    currdim = ("sst_x", "sst_y")
                else:
                    currdim = "t"
                datasets[v]["dims"] = currdim  # this sets the dimensions for v

            utils.write_netcdf(output_file, datasets, dims, atts=global_atts)
            # sys.exit()


if __name__ == "__main__":
    main()
