# from __future__ import absolute_import, division, print_function, unicode_literals

import matplotlib

matplotlib.use("Agg")

import os

pltg = os.environ["PLTG"]
import numpy as np
import sys

import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))

import argparse
import glob
from datetime import datetime, timedelta, timezone
import time
import pickle
import matplotlib.pyplot as plt
import collections

parser = argparse.ArgumentParser()
parser.add_argument(
    "-i",
    "--indir",
    help="Directory containing trained CNN model and config.",
    type=str,
    nargs=1,
)
parser.add_argument(
    "-o",
    "--outdir",
    help="Output directory for model and figures. Default=args.indir + '/TEST/'",
    nargs=1,
    type=str,
)
parser.add_argument(
    "-m",
    "--model",
    help="Use this model instead of fit_conv_model.h5 in --indir.",
    type=str,
    nargs=1,
)
parser.add_argument(
    "-n",
    "--ncpus",
    help="Limit TF to this number of CPUs. Default is to use all available.",
    type=int,
    default=[-1],
    nargs=1,
)

args = parser.parse_args()

indir = args.indir[0]

config = pickle.load(open(os.path.join(indir, "model_config.pkl"), "rb"))
model_file = args.model[0] if (args.model) else f"{indir}/fit_conv_model.h5"

outdir = args.outdir[0] if (args.outdir) else os.path.join(indir, "TEST")


ncpus = args.ncpus[0]

INPUTS = config["inputs"]
try:
    BINARIZE = config["binarize1"]
except KeyError:
    BINARIZE = 1
try:
    BINARIZE2 = config["binarize2"]
except KeyError:
    pass
try:
    TARGET1 = config["target1"]
except KeyError:
    TARGET1 = config["target"]
try:
    TARGET2 = config["target2"]
except KeyError:
    TARGET2 = None
NGPU = 2  # config['ngpu']
batchsize = 16
assert batchsize % NGPU == 0
if NGPU == 0:
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
# else:
#  os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID"
#  os.environ["CUDA_VISIBLE_DEVICES"]="2" #hard-coded!
import tensorflow as tf

tf.keras.backend.clear_session()
from lightningcast import losses
from lightningcast import tf_metrics
import tensorflow_addons as tfa

optimizer = (
    tfa.optimizers.AdaBelief(learning_rate=0.0001)
    if (config["optimizer"].lower() == "adabelief")
    else "Adam"
)

if ncpus > 0:
    # limiting CPUs
    num_threads = 30
    os.environ["OMP_NUM_THREADS"] = str(num_threads)
    os.environ["TF_NUM_INTRAOP_THREADS"] = str(num_threads)
    os.environ["TF_NUM_INTEROP_THREADS"] = str(num_threads)
    tf.config.threading.set_inter_op_parallelism_threads(num_threads)
    tf.config.threading.set_intra_op_parallelism_threads(num_threads)

from tensorflow import keras

gpus = tf.config.list_physical_devices("GPU")
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)
from tensorflow.keras.models import load_model

# ------------------------------------------------------------------------------------------------------------------------------------------------------
def make_divisible(X, batch_size, n_gpu):
    """
    This function takes our batch_size and n_gpu we're using and
    will create a new number of samples X that is divisible by both.
    We will then use dataset.take(amount_to_add) to replicate a few
    samples. We do this so that we don't have any remainders in our
    dataset during evaluation.
    """

    amount_to_add = 0
    while X % batch_size != 0 or X % n_gpu != 0:
        X += 1
        amount_to_add += 1

    assert X % batch_size == 0 and X % n_gpu == 0

    return X, amount_to_add


# ---------------------------------------------------------------------------------------------------------------------------------------------------
def get_metrics():

    metrics = [
        "accuracy",
        tf.keras.metrics.AUC(name="aupr", curve="PR"),
        tf.keras.metrics.MeanSquaredError(name="brier_score"),
    ]

    custom_metrics = collections.OrderedDict()
    custom_metrics["loss"] = "foo"
    custom_metrics["accuracy"] = "foo"
    custom_metrics["aupr"] = tf.keras.metrics.AUC(name="aupr", curve="PR")
    custom_metrics["brier_score"] = tf.keras.metrics.MeanSquaredError(name="brier_score")

    if TARGET2:
        ntargets = 2
    else:
        ntargets = 1

    for jj in range(ntargets):
        for ii in np.arange(0.05,1,0.05):
    
            met_name = f"csi{str(int(round(ii*100))).zfill(2)}_index{jj}"
            csi_met = losses.csi(
                use_as_loss_function=False,
                use_soft_discretization=False,
                hard_discretization_threshold=ii,
                name=met_name,
                index=jj
            )
            metrics.append(csi_met)
            custom_metrics[met_name] = csi_met
    
            met_name = f"pod{str(int(round(ii*100))).zfill(2)}_index{jj}"
            pod_met = tf_metrics.pod(
                use_soft_discretization=False, hard_discretization_threshold=ii, name=met_name, index=jj
            )
            metrics.append(pod_met)
            custom_metrics[met_name] = pod_met
    
            met_name = f"far{str(int(round(ii*100))).zfill(2)}_index{jj}"
            far_met = tf_metrics.far(
                use_soft_discretization=False, hard_discretization_threshold=ii, name=met_name, index=jj
            )
            metrics.append(far_met)
            custom_metrics[met_name] = far_met

            met_name = f"obsct{str(int(round(ii*100))).zfill(2)}_index{jj}"
            obsct_met = tf_metrics.obs_ct(
                threshold1=ii-0.05, threshold2=ii, name=met_name, index=jj
            )
            metrics.append(obsct_met)
            custom_metrics[met_name] = obsct_met

            met_name = f"fcstct{str(int(round(ii*100))).zfill(2)}_index{jj}"
            fcstct_met = tf_metrics.fcst_ct(
                threshold1=ii-0.05, threshold2=ii, name=met_name, index=jj
            )
            metrics.append(fcstct_met)
            custom_metrics[met_name] = fcstct_met


    return custom_metrics, metrics

# ----------------------------------------------------------------------------------------------------------------
def parse_tfrecord_control(example):

    feature_description = {
        "CH02": tf.io.FixedLenFeature([], tf.string),
        "CH05": tf.io.FixedLenFeature([], tf.string),
        "CH13": tf.io.FixedLenFeature([], tf.string),
        "CH15": tf.io.FixedLenFeature([], tf.string),
        "FED_accum_60min_2km": tf.io.FixedLenFeature([], tf.string),
    }
    features = tf.io.parse_single_example(example, feature_description)

    ch02 = tf.reshape(tf.io.parse_tensor(features["CH02"], tf.uint8), [1920, 1920, 1])
    # Unscaling the TensorFlow tensor to native units. Then applying mean/std scaling, since the model was trained thusly.
    ch02 = tf.cast(ch02, dtype=tf.float32)
    vmax = tf.constant(1.0, dtype=tf.float32)  # Assuming vmax and vmin are constants
    vmin = tf.constant(
        0.0, dtype=tf.float32
    )  # These are values for how the tfrecords were scaled
    two56 = tf.constant(255.9999, dtype=tf.float32)
    ch02 = ch02 / two56 * (vmax - vmin) + vmin
    mean = tf.constant(0.08713, dtype=tf.float32)
    std = tf.constant(0.15314, dtype=tf.float32)
    features["input_1"] = (ch02 - mean) / std

    ch05 = tf.reshape(tf.io.parse_tensor(features["CH05"], tf.uint8), [960, 960, 1])
    # Unscaling the TensorFlow tensor to native units. Then applying mean/std scaling, since the model was trained thusly.
    ch05 = tf.cast(ch05, dtype=tf.float32)
    vmax = tf.constant(0.75, dtype=tf.float32)  # Assuming vmax and vmin are constants
    vmin = tf.constant(
        0.0, dtype=tf.float32
    )  # These are values for how the tfrecords were scaled
    ch05 = ch05 / two56 * (vmax - vmin) + vmin
    mean = tf.constant(0.047332, dtype=tf.float32)
    std = tf.constant(0.0834376, dtype=tf.float32)
    features["input_2"] = (ch05 - mean) / std

    ch13 = tf.reshape(tf.io.parse_tensor(features["CH13"], tf.uint8), [480, 480, 1])
    # Unscaling the TensorFlow tensor to native units. Then applying mean/std scaling, since the model was trained thusly.
    ch13 = tf.cast(ch13, dtype=tf.float32)
    vmax = tf.constant(320, dtype=tf.float32)  # Assuming vmax and vmin are constants
    vmin = tf.constant(
        190, dtype=tf.float32
    )  # These are values for how the tfrecords were scaled
    ch13 = ch13 / two56 * (vmax - vmin) + vmin
    mean = tf.constant(270.93, dtype=tf.float32)
    std = tf.constant(25.9126, dtype=tf.float32)
    ch13 = (ch13 - mean) / std

    ch15 = tf.reshape(tf.io.parse_tensor(features["CH15"], tf.uint8), [480, 480, 1])
    # Unscaling the TensorFlow tensor to native units. Then applying mean/std scaling, since the model was trained thusly.
    ch15 = tf.cast(ch15, dtype=tf.float32)
    vmax = tf.constant(320, dtype=tf.float32)  # Assuming vmax and vmin are constants
    vmin = tf.constant(
        190, dtype=tf.float32
    )  # These are values for how the tfrecords were scaled
    ch15 = ch15 / two56 * (vmax - vmin) + vmin
    mean = tf.constant(273.168, dtype=tf.float32)
    std = tf.constant(19.9895, dtype=tf.float32)
    ch15 = (ch15 - mean) / std

    features["input_3"] = tf.concat([ch13, ch15], axis=-1)

    fed_accum = tf.reshape(
        tf.io.parse_tensor(features["FED_accum_60min_2km"], tf.uint8), [480, 480]
    )  # no last dim here

    # binarize
    zero = tf.zeros_like(fed_accum)
    one = tf.ones_like(fed_accum)
    targets = tf.where(fed_accum >= BINARIZE, x=one, y=zero)

    return (features, targets)


# ----------------------------------------------------------------------------------------------------------------
def parse_tfrecord_fn(example):

    feature_description = {
        "CH02": tf.io.FixedLenFeature([], tf.string),
        "CH05": tf.io.FixedLenFeature([], tf.string),
        "CH08": tf.io.FixedLenFeature([], tf.string),
        "CH09": tf.io.FixedLenFeature([], tf.string),
        "CH10": tf.io.FixedLenFeature([], tf.string),
        "CH11": tf.io.FixedLenFeature([], tf.string),
        "CH13": tf.io.FixedLenFeature([], tf.string),
        "CH15": tf.io.FixedLenFeature([], tf.string),
        "FED_accum_60min_2km": tf.io.FixedLenFeature([], tf.string),
    }
    features = tf.io.parse_single_example(example, feature_description)
    features["input_0"] = tf.reshape(
        tf.io.parse_tensor(features["CH02"], tf.uint8), [1920, 1920, 1]
    )
    features["input_1"] = tf.reshape(
        tf.io.parse_tensor(features["CH05"], tf.uint8), [960, 960, 1]
    )
    # ch08 = tf.reshape(tf.io.parse_tensor(features["CH08"], tf.uint8), [480,480,1])
    #  ch09 = tf.reshape(tf.io.parse_tensor(features["CH09"], tf.uint8), [480,480,1])
    #  ch10 = tf.reshape(tf.io.parse_tensor(features["CH10"], tf.uint8), [480,480,1])
    # ch11 = tf.reshape(tf.io.parse_tensor(features["CH11"], tf.uint8), [480,480,1])

    for chan in INPUTS[-1]:  # 2-km input
        data_2km = tf.reshape(
            tf.io.parse_tensor(features[chan], tf.uint8), [480, 480, 1]
        )
        if "input_2" in features:
            features["input_2"] = tf.concat([features["input_2"], data_2km], axis=-1)
        else:
            features["input_2"] = data_2km

    # ch13 = tf.reshape(tf.io.parse_tensor(features["CH13"], tf.uint8), [480,480,1])
    # ch15 = tf.reshape(tf.io.parse_tensor(features["CH15"], tf.uint8), [480,480,1])
    # features['input_2'] = tf.concat([ch13, ch15], axis=-1)

    # binarize
    fed_accum = tf.reshape(
        tf.io.parse_tensor(features["FED_accum_60min_2km"], tf.uint8), [480, 480, 1]
    )
    zero = tf.zeros_like(fed_accum)
    one = tf.ones_like(fed_accum)
    targets = tf.where(fed_accum >= BINARIZE, x=one, y=zero)
    if TARGET2:
        # a second target threshold....assuming target1 and target2 are the same here, but binarize2 > binarize1
        targets2 = tf.where(fed_accum >= BINARIZE2, x=one, y=zero)
        targets = tf.concat([targets, targets2], axis=-1)

    return (features, targets)


################################################################################################################

# Load model and set up GPUs

if NGPU > 1:
    # Create a MirroredStrategy.
    strategy = tf.distribute.MirroredStrategy()
    print("Number of devices: {}".format(strategy.num_replicas_in_sync))
    with strategy.scope():
        custom_objs, metrics = get_metrics()
        conv_model = load_model(model_file, custom_objects=custom_objs, compile=False)
        conv_model.compile(
            optimizer=optimizer, loss=config["loss_fcn"], metrics=metrics
        )
else:
    custom_objs, metrics = get_metrics()
    conv_model = load_model(model_file, custom_objects=custom_objs, compile=False)
    conv_model.compile(optimizer=optimizer, loss=config["loss_fcn"], metrics=metrics)

print(conv_model.summary())


AUTOTUNE = tf.data.AUTOTUNE

# Assuming 'static' means control model (LCv1)
parse_fn = parse_tfrecord_control if ("static" in model_file) else parse_tfrecord_fn

# Create the test dataset for running the permutation
sat = "goes16"
year = "2021"
inroot = f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/{year}/"

# leave empty if you only want one run
subdirs = []  #'01-02','03','04','05','06','07','08','09','10','11-12']

if len(subdirs) > 0:
    val_lists = []
    for sub in subdirs:
        if sub == "01-02":
            val_lists.append(f"{inroot}/{year}0[1-2]*/*tfrec")
        elif sub == "11-12":
            val_lists.append(f"{inroot}/{year}1[1-2]*/*tfrec")
        else:
            val_lists.append(f"{inroot}/{year}{sub}*/*tfrec")
else:
    val_lists = [f"{inroot}/{year}*/*tfrec"]  # this is your "one-run" case


for ii, val_list in enumerate(val_lists):
    test_filenames = sorted(glob.glob(val_list))  # sort to ensure reproducibility
    n_samples = len(test_filenames)
    print("\nNumber of samples:", n_samples, "\n")

    # Create the TFRecordDataset
    test_ds = tf.data.TFRecordDataset(test_filenames, num_parallel_reads=AUTOTUNE)
    # Get remainder dataset
    _, amt_to_add = make_divisible(n_samples, batchsize, NGPU)
    assert (
        amt_to_add / n_samples < 0.01
    )  # make sure we're not duplicating too many samples
    remainder_ds = test_ds.take(amt_to_add)
    # Concatenate to test_ds
    test_ds = test_ds.concatenate(remainder_ds)

    test_ds = (
        test_ds.map(parse_fn, num_parallel_calls=AUTOTUNE)
        .batch(batchsize)  # Don't batch if making numpy arrays
        .prefetch(AUTOTUNE)
    )

    # Disable AutoShard.
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = (
        tf.data.experimental.AutoShardPolicy.OFF
    )
    test_ds = test_ds.with_options(options)

    # Perform eval
    eval_results = conv_model.evaluate(test_ds)

    cter = 0
    dict_results = collections.OrderedDict(
        {
            "n_gpu": NGPU,
            "batch_size": batchsize,
            "n_samples": n_samples,
            "n_duplicates": amt_to_add,
        }
    )
    for key in custom_objs:
        dict_results[key] = np.round(eval_results[cter], 5)
        cter += 1

    # Append outsubdir, if necessary
    try:
        final_outdir = f"{outdir}/month{subdirs[ii]}"
    except IndexError:
        final_outdir = outdir

    os.makedirs(final_outdir, exist_ok=True)
    pickle.dump(dict_results, open(f"{final_outdir}/eval_results.pkl", "wb"))
    print(f"Saved {final_outdir}/eval_results.pkl")


# ny,nx = config['ps_2km']
#
# val_list = config['val_list']
#
# val_files = np.sort(glob.glob(val_list)) #val_list is a wildcard string
##According to https://www.tensorflow.org/api_docs/python/tf/data/Dataset#list_files ,
## If your filenames have already been globbed, use Dataset.from_tensor_slices(filenames) instead,
## as re-globbing every filename with list_files may result in poor performance with remote storage systems.
# val_ds = tf.data.Dataset.from_tensor_slices(val_files) #N.B.: val_ds and val_files must have same order for spatial verification
# val_ds = val_ds.map(parse_data, num_parallel_calls=tf.data.experimental.AUTOTUNE)
# n_vsamples = len(val_ds); config['n_vsamples'] = n_vsamples
# print('#validation samples:',n_vsamples)
# val_bs = 1000
# val_ds = val_ds.batch(val_bs)
# val_ds = val_ds.prefetch(tf.data.experimental.AUTOTUNE)
## Change AutoShard policy. Not exactly sure what this does, but it at least gets rid of annoying warning messages.
# options = tf.data.Options()
# options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
# val_ds = val_ds.with_options(options)
#
##preds = conv_model.predict(val_ds)
##print(preds.shape)
##print(preds.dtype)
##print('GB:',preds.nbytes)/1024**3
#
##preallocate
# all_preds = np.empty((n_vsamples,ny//stride,nx//stride),dtype=np.float16) #Striding to save space and time.
# all_labels = np.zeros((n_vsamples,ny//stride,nx//stride),dtype=np.byte)
##float32 precision needed for validation routines (e.g., MSE)
#
# cter = 0
# for inputs,labels in val_ds:
#  t0=time.time()
#  preds = conv_model.predict(inputs)
#  print(time.time()-t0)
#  all_preds[cter:cter+val_bs] = preds[:,::stride,::stride]
#  all_labels[cter:cter+val_bs] = labels[:,::stride,::stride,0]
#  cter+=val_bs
#
# basenames = [os.path.basename(vf) for vf in val_files] #20190607-2301_Y192-X1984.bmp
# datetimes = [datetime.strptime(x.split('_')[0],'%Y%m%d-%H%M') for x in basenames]
# midptY = [int(x.split('_Y')[1].split('-X')[0]) for x in basenames]
# midptX = [int(x.split('-X')[1].split('.')[0]) for x in basenames]
# strideY = np.array(midptY) // stride
# strideX = np.array(midptX) // stride
#
##save the predictions / labels first
# preds_labs_dict1 = {'preds':all_preds,'labels':all_labels,'files':val_files,
#                    'datetimes':datetimes,
#                    'midptX':midptX, 'midptY':midptY,
#                    'strideX':strideX, 'strideY':strideY, 'stride':stride}
# pickle.dump(preds_labs_dict1, open(f"{outdir}/all_preds_labs.pkl","wb"))
