from __future__ import absolute_import, division, print_function, unicode_literals

import matplotlib

matplotlib.use("Agg")
import os, sys

# os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID";
# os.environ["CUDA_VISIBLE_DEVICES"]="2";

import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))

import shutil
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from lightningcast import performance_diagrams
from lightningcast import roc_curves
from lightningcast import attributes_diagrams
import pickle
import glob
import copy

import tensorflow as tf

tf.keras.backend.clear_session()
# from tensorflow import keras
# import tensorflow.keras.backend as K
# from tensorflow.keras import mixed_precision
# set the mixed precision policy
# mixed_precision.set_global_policy('mixed_float16')

from lightningcast import tf_models
from tensorflow.keras.models import Model, save_model, load_model
from tensorflow.keras.callbacks import (
    EarlyStopping,
    ModelCheckpoint,
    ReduceLROnPlateau,
    CSVLogger,
    TensorBoard,
)
from sklearn.metrics import mean_squared_error, roc_auc_score
from sklearn.metrics import auc as skl_auc
from datetime import datetime, timedelta
import time
import argparse
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
from lightningcast.LRFinder import LRFinder

from lightningcast import losses

# gpus = tf.config.list_physical_devices('GPU')
# for gpu in gpus:
#  tf.config.experimental.set_memory_growth(gpu, True)

parser = argparse.ArgumentParser()
parser.add_argument(
    "-m", "--model", help="import this trained CNN model", type=str, nargs=1
)
parser.add_argument(
    "-c",
    "--config_file",
    help="enter a config pkl file (usu. from a previous training session)",
    nargs=1,
    type=str,
)

args = parser.parse_args()


if not (args.config_file):
    from lightningcast import tf_config

    config = tf_config.tf_config()
else:
    config = pickle.load(open(args.config_file[0], "rb"))

# some global vars
INPUTS = config["inputs"]
BINARIZE1 = config["binarize1"]
BINARIZE2 = config["binarize2"]
TARGET1 = config["target1"]
TARGET2 = config["target2"]
NGPU = config["ngpu"]
BATCHSIZE = config["batchsize"]
LOSS_FCN = config["loss_fcn"]
POS_WEIGHT = config["pos_weight"] # Weight for positive class (1 is default)
                                  # 0 means this is not used (i.e., weight of 1 everywhere)
try:
    STD = config["img_aug"]["random_noise"]
except KeyError:
    pass

byte_scaling_vals = config["byte_scaling_vals"]
# -------------------------------------------------------------------------------
def get_best_model(outdir):

    best_model = "foo"
    best_loss = 999999

    models = glob.glob(f"{outdir}/model-*.h5")
    for mod in models:
        loss = float(os.path.basename(mod).split("-")[-1].split(".h5")[0])
        if loss < best_loss:
            best_loss = loss
            best_model = mod

    if best_model == "foo":
        logging.warning("Couldn't find best model. Returning last model.")
        try:
            return np.sort(models)[-1]
        except IndexError:
            raise
    else:
        return best_model


##################################################################################################################
def training_history_figs(history, outdir):

    # summarize history for accuracy
    try:
        plt.plot(history["binary_accuracy"])
        if "val_binary_accuracy" in history:
            plt.plot(history["val_binary_accuracy"])
        plt.title("model binary accuracy")
        plt.ylabel("accuracy")
        plt.xlabel("epoch")
        if "val_binary_accuracy" in history:
            plt.legend(["train", "test"], loc="upper left")
        else:
            plt.legend(["train"], loc="upper left")
        plt.savefig(os.path.join(outdir, "binary_accuracy_history.png"))
        plt.close()
    except KeyError:
        plt.plot(history["accuracy"])
        if "val_accuracy" in history:
            plt.plot(history["val_accuracy"])
        plt.title("model accuracy")
        plt.ylabel("accuracy")
        plt.xlabel("epoch")
        if "val_accuracy" in history:
            plt.legend(["train", "test"], loc="upper left")
        else:
            plt.legend(["train"], loc="upper left")
        plt.savefig(os.path.join(outdir, "accuracy_history.png"))
        plt.close()

    # summarize history for loss
    plt.plot(history["loss"])
    if "val_loss" in history:
        plt.plot(history["val_loss"])
    plt.title("model loss")
    plt.ylabel("loss")
    plt.xlabel("epoch")
    if "val_loss" in history:
        plt.legend(["train", "test"], loc="upper left")
    else:
        plt.legend(["train"], loc="upper left")
    plt.savefig(os.path.join(outdir, "loss_history.png"))
    plt.close()


##################################################################################################################
def add_scores_to_file(outdir):
    """
    Will append to {outdir}/verification_scores.txt or create it if it does not exist.
    If scores['score']['best_score'] = 0, then "higher_is_better" = True.
    If scores['score']['best_score'] = 1, then "higher_is_better" = False (i.e., lower is better).
    """

    of = open(f"{outdir}/log.csv", "r")
    lines = of.readlines()
    of.close()

    scores = {}

    for ii, line in enumerate(lines):
        parts = line.split(",")
        if ii == 0:  # first line
            for score in parts:
                if score.startswith('val_'):
                    scores[score] = {'idx': parts.index(score)}
                    scores[score]["best_epoch"] = 0
                    if 'csi' in score or 'pod' in score or 'aupr' in score or 'accuracy' in score:
                        scores[score]["higher_is_better"] = True
                        scores[score]["best_score"] = 0
                    elif 'far' in score or 'brier_score' in score or 'loss' in score:
                        scores[score]["higher_is_better"] = False 
                        scores[score]["best_score"] = 1

        else:  # All other lines
            for score in scores:
                idx = scores[score]["idx"]
                best_val_score = scores[score]["best_score"]
                val_score = float(parts[idx])  # the validation metric for this epoch
                if scores[score]["higher_is_better"]:
                    if val_score > best_val_score:
                        scores[score]["best_score"] = val_score
                        scores[score]["best_epoch"] = int(parts[0]) + 1
                else:  # lower is better
                    if val_score < best_val_score:
                        scores[score]["best_score"] = val_score
                        scores[score]["best_epoch"] = int(parts[0]) + 1

    of = open(f"{outdir}/verification_scores.txt", "a")
    for score in scores:
        best_score = scores[score]["best_score"]
        best_epoch = scores[score]["best_epoch"]
        of.write(f"Best {score}: {np.round(best_score,5)}; at epoch {best_epoch}\n")
    of.close()


##################################################################################################################
def parse_tfrecord_fn(example):

    feature_description = {
        "CH02": tf.io.FixedLenFeature([], tf.string),
        "CH05": tf.io.FixedLenFeature([], tf.string),
        "CH08": tf.io.FixedLenFeature([], tf.string),
        "CH09": tf.io.FixedLenFeature([], tf.string),
        "CH10": tf.io.FixedLenFeature([], tf.string),
        "CH11": tf.io.FixedLenFeature([], tf.string),
        "CH13": tf.io.FixedLenFeature([], tf.string),
        "CH15": tf.io.FixedLenFeature([], tf.string),
        # "SATZEN": tf.io.FixedLenFeature([], tf.string),
        # "SFC_ELEV": tf.io.FixedLenFeature([], tf.string),
        # "CLIMO": tf.io.FixedLenFeature([], tf.string),
        # "REF10": tf.io.FixedLenFeature([], tf.string),
        # "FED": tf.io.FixedLenFeature([], tf.string),
        "FED_accum_60min_2km": tf.io.FixedLenFeature([], tf.string),
        "FED_accum_30min_2km": tf.io.FixedLenFeature([], tf.string),
    }
    features = tf.io.parse_single_example(example, feature_description)
    # features["image"] = tf.io.decode_jpeg(example["FED_accum_15min_2km"], channels=1)
    features["CH02"] = tf.reshape(
        tf.io.parse_tensor(features["CH02"], tf.uint8), [1920, 1920, 1]
    )
    features["CH05"] = tf.reshape(
        tf.io.parse_tensor(features["CH05"], tf.uint8), [960, 960, 1]
    )
    # features["REF10"] = tf.reshape(tf.io.parse_tensor(features["REF10"], tf.uint8), [960,960,1])
    features["CH08"] = tf.reshape(
        tf.io.parse_tensor(features["CH08"], tf.uint8), [480, 480, 1]
    )
    features["CH09"] = tf.reshape(
        tf.io.parse_tensor(features["CH09"], tf.uint8), [480, 480, 1]
    )
    features["CH10"] = tf.reshape(
        tf.io.parse_tensor(features["CH10"], tf.uint8), [480, 480, 1]
    )
    features["CH11"] = tf.reshape(tf.io.parse_tensor(features["CH11"], tf.uint8), [480,480,1])
    features["CH13"] = tf.reshape(
        tf.io.parse_tensor(features["CH13"], tf.uint8), [480, 480, 1]
    )
    features["CH15"] = tf.reshape(
        tf.io.parse_tensor(features["CH15"], tf.uint8), [480, 480, 1]
    )
    # features["CLIMO"] = tf.reshape(tf.io.parse_tensor(features["CLIMO"], tf.uint8), [120,120,1])
    # features["SFC_ELEV"] = tf.reshape(tf.io.parse_tensor(features["SFC_ELEV"], tf.uint8), [480,480,1])
    # features["FED"] = tf.reshape(tf.io.parse_tensor(features["FED"], tf.uint8), [480,480,1])
    # features["FED_accum_15min_2km"] = tf.reshape(tf.io.parse_tensor(features["FED_accum_15min_2km"], tf.uint8), [480,480,1])
    features["FED_accum_60min_2km"] = tf.reshape(
        tf.io.parse_tensor(features["FED_accum_60min_2km"], tf.uint8), [480, 480, 1]
    )
    features["FED_accum_30min_2km"] = tf.reshape(
        tf.io.parse_tensor(features["FED_accum_30min_2km"], tf.uint8), [480, 480, 1]
    )  
 
    return features


##################################################################################################################
def prepare_sample(features):

    # binarize
    targetImage_int = features[TARGET1]
    zero = tf.zeros_like(targetImage_int)
    one = tf.ones_like(targetImage_int)
    if LOSS_FCN == "QR":  # don't binarize if doing quantile regression!
        targetImage = targetImage_int
    else:  # normal classification
        targetImage = tf.where(targetImage_int >= BINARIZE1, x=one, y=zero)
        if TARGET2:  # a second target threshold
            targetImage_int = features[TARGET2]
            zero = tf.zeros_like(targetImage_int)
            one = tf.ones_like(targetImage_int)
            targetImage2 = tf.where(targetImage_int >= BINARIZE2, x=one, y=zero)
            targetImage = tf.concat([targetImage, targetImage2], axis=2)

    # Read the related *predictor* image .bmp files
    inputs = {}

    inputs["input_0"] = features["CH02"]
    inputs["input_1"] = features[
        "CH05"
    ]  # tf.concat([features['CH05'],features['REF10']], axis=2)
    ch13 = features["CH13"]
    ch11 = features['CH11']
    ch08 = features['CH08']
    ch09 = features['CH09']
    ch10 = features['CH10']
    ch15 = features["CH15"]
    # satzen = features['SATZEN']
    # sfc_elev = features['SFC_ELEV']
    # fed = features['FED']
    inputs["input_2"] = tf.concat([ch13, ch15], axis=2)
    # inputs['input_3'] = features['CLIMO']

    if POS_WEIGHT > 0:
        class_weight = tf.multiply(one, POS_WEIGHT)
        sample_weights = tf.where(targetImage_int >= BINARIZE1, x=class_weight, y=one)
        return inputs, targetImage, sample_weights
    else:
        return inputs, targetImage


#################################################################################################################################
def apply_augmentations(input_dict, target_img):
    """Apply data augmentations

    Args:
        x: input_dict, target_img

    Returns:
        augmented input_dict and target_img
    """

    # See https://benihime91.github.io/blog/deeplearning/tensorflow2.x/data_augmentation/image/2020/11/15/image-augmentation-tensorflow.html

    # randomly rotates an image by a factor of 90 degree
    rotate_prob = tf.random.uniform([], 0, 1.0, dtype=tf.float32)

    # We often have multiple TF/Keras Inputs in the model
    for (
        key,
        inp,
    ) in (
        input_dict.items()
    ):  # N.B. for some reason, we can't put for-loops within the if/elif/else block.
        if rotate_prob > 0.66:
            inp = tf.image.rot90(inp, k=3)
            krot = 3  # rotate 270º
        elif rotate_prob > 0.33:
            inp = tf.image.rot90(inp, k=2)
            krot = 2  # rotate 180º
        else:
            inp = tf.image.rot90(inp, k=1)
            krot = 1  # rotate 90º
        input_dict[key] = inp

    target_img = tf.image.rot90(target_img, k=krot)

    return input_dict, target_img


#################################################################################################################################
def apply_noise(input_dict, target_img):
    """Apply random Gaussian noise augmentation

    Args:
        x: input_dict, target_img

    Returns:
        augmented input_dict and target_img
    """

    # See https://stackoverflow.com/questions/41174769/additive-gaussian-noise-in-tensorflow
    # Amended for tf.uint8 data
    # STD is global variable above
    for key, inp in input_dict.items():

        # Convert uint8 to float32
        inp_float = tf.cast(inp, tf.float32)

        # Apply random Gaussian noise
        noise = tf.random.normal(
            shape=tf.shape(inp_float), mean=0, stddev=STD, dtype=tf.float32
        )

        # Add noise and clip to ensure values stay within uint8 range
        inp_with_noise_float = tf.clip_by_value(inp_float + noise, 0.0, 255.0)

        # Convert back to uint8
        input_dict[key] = tf.cast(inp_with_noise_float, tf.uint8)

    return input_dict, target_img


#################################################################################################################################
# Train: Create list of training datasets
print("\nBuilding training Dataset")

AUTOTUNE = tf.data.AUTOTUNE

train_filenames = tf.io.gfile.glob(config["train_list"])
n_tsamples = len(train_filenames)
train_ds = (
    tf.data.TFRecordDataset(train_filenames, num_parallel_reads=AUTOTUNE)
    .shuffle(1024)
    .map(parse_tfrecord_fn, num_parallel_calls=AUTOTUNE)
    .map(prepare_sample, num_parallel_calls=AUTOTUNE)
    .batch(BATCHSIZE)
    .prefetch(AUTOTUNE)
)

# Augment data (optional)
if "random_rotation" in config["img_aug"]:
    aug_ds = train_ds.map(apply_augmentations, num_parallel_calls=AUTOTUNE)
    # Concatenate the original and augmented datasets (doubles dataset)
    train_ds = train_ds.concatenate(aug_ds)
    del aug_ds
if (
    "random_noise" in config["img_aug"]
):  # original and concatenated augmentations are noised
    train_ds = train_ds.map(apply_noise, num_parallel_calls=AUTOTUNE)

# for quick_looks
# for inputs,labels in train_ds:
#  inp = inputs['input_0'].numpy()
#  #inp2 = inputs['input_1'].numpy()
#  labs = labels.numpy()
#  fig,(ax1,ax2,ax3) = plt.subplots(nrows=1,ncols=3)
#  ax1.imshow(inp[...,0])
#  ax2.imshow(inp[...,1])
#  ax3.imshow(labs[...,0])
#  plt.show()
#  sys.exit()

print("built training dataset")

print("\nBuilding validation Dataset")

outdir = config["outdir"]
os.makedirs(outdir, exist_ok=True)

# Change AutoShard policy. Not exactly sure what this does, but it at least gets rid of annoying warning messages.
options = tf.data.Options()
options.experimental_distribute.auto_shard_policy = (
    tf.data.experimental.AutoShardPolicy.DATA
)
train_ds = train_ds.with_options(options)

val_filenames = tf.io.gfile.glob(config["val_list"])
n_vsamples = len(val_filenames)
val_ds = (
    tf.data.TFRecordDataset(val_filenames, num_parallel_reads=AUTOTUNE)
    .map(parse_tfrecord_fn, num_parallel_calls=AUTOTUNE)
    .map(prepare_sample, num_parallel_calls=AUTOTUNE)
    .batch(BATCHSIZE)
    .prefetch(AUTOTUNE)
)
val_ds = val_ds.with_options(options)

# print out config options
ofile = os.path.join(outdir, "model_config.txt")
of = open(ofile, "w")
for key, value in config.items():
    of.write(str(key) + ": " + str(value) + "\n")
of.write(f"Number of training samples: {n_tsamples}\n")
of.write(f"Number of validation samples: {n_vsamples}\n")
of.close()
pickle.dump(config, open(os.path.join(outdir, "model_config.pkl"), "wb"))

# ===================================================================================================================#

csvlogger = CSVLogger(f"{outdir}/log.csv", append=True)
early_stopping = EarlyStopping(
    monitor="val_loss", patience=config["es_patience"], min_delta=0.00001
)
mcp_save = ModelCheckpoint(
    os.path.join(outdir, "model-{epoch:02d}-{val_loss:03f}.h5"),
    save_best_only=True,
    monitor="val_loss",
    mode="min",
)
reduce_lr_loss = ReduceLROnPlateau(
    monitor="val_loss",
    cooldown=config["rlr_cooldown"],
    verbose=1,
    min_delta=0.00001,
    factor=config["rlr_factor"],
    patience=config["rlr_patience"],
    mode="min",
)
tboard = TensorBoard(
    log_dir="/home/jcintineo/lightningcast/src/tb_logs/",
    histogram_freq=1,
    write_images=True,
    profile_batch=(40, 60),
)  # profile from batch X to batch Y
# lr_finder = LRFinder(min_lr=1e-7, max_lr=1e-1)
# callbacks = [lr_finder]
callbacks = [early_stopping, mcp_save, reduce_lr_loss, csvlogger]  # , tboard]


if NGPU > 1:
    # Create a MirroredStrategy.
    strategy = tf.distribute.MirroredStrategy()
    print("Number of devices: {}".format(strategy.num_replicas_in_sync))
    with strategy.scope():
        if args.model:
            logging.info("Opening " + args.model[0])
            CUSTOM_METRICS, METRICS = tf_models.get_metrics()
            conv_model = load_model(args.model[0], custom_objects=CUSTOM_METRICS)
            conv_model.compile(
                loss=conv_model.loss, optimizer=conv_model.optimizer, metrics=METRICS
            )
            print(conv_model.summary())
            initial_epoch = int(os.path.basename(args.model[0]).split("-")[1])
        else:
            tf_model = getattr(tf_models, config["cnn"])
            conv_model = tf_model(config=config)
            initial_epoch = 0
else:
    if args.model:
        logging.info("Opening " + args.model[0])
        CUSTOM_METRICS, METRICS = tf_models.get_metrics()
        conv_model = load_model(args.model[0], custom_objects=CUSTOM_METRICS)
        conv_model.compile(
            loss=conv_model.loss, optimizer=conv_model.optimizer, metrics=METRICS
        )
        print(conv_model.summary())
        initial_epoch = int(os.path.basename(args.model[0]).split("-")[1])
    else:
        tf_model = getattr(tf_models, config["cnn"])
        conv_model = tf_model(config=config)
        initial_epoch = 0
# sys.exit()
# -----------------------------------------------------------------------------------------------------
print("input shapes:", config["input_tuples"])
logging.info("Fitting model...")

# FIT THE MODEL
history = conv_model.fit(
    x=train_ds,
    verbose=1,
    epochs=config["nepoch"],
    validation_data=val_ds,
    initial_epoch=initial_epoch,
    callbacks=callbacks,
)
# shuffle has no effect if generator OR a tf.data.Dataset

logging.info("Saving model and training history...")
# copy the model with the best val_loss
try:
    best_model = get_best_model(outdir)
except IndexError:
    logging.error("Something is wrong with finding the best model. Exiting.")
    sys.exit()
else:
    shutil.copy(best_model, f"{os.path.dirname(best_model)}/fit_conv_model.h5")

# load the best-val_loss model
conv_model = load_model(
    f"{outdir}/fit_conv_model.h5", compile=False
)  # False b/c no more training

# It seems like 1 GPU is just as fast as 2 GPUs, so commented this out for now.
# load the best-val_loss model
# if(NGPU > 1):
#  # Create a MirroredStrategy.
#  strategy = tf.distribute.MirroredStrategy()
#  print("Number of devices: {}".format(strategy.num_replicas_in_sync))
#  with strategy.scope():
#      conv_model = load_model(f"{outdir}/fit_conv_model.h5",compile=False) #False b/c no more training
# else:
#  conv_model = load_model(f"{outdir}/fit_conv_model.h5",compile=False) #False b/c no more training

pickle.dump(
    history.history,
    open(os.path.join(outdir, "training_history.pkl"), "wb"),
    protocol=4,
)
training_history_figs(history.history, outdir)

################################################
# sys.exit()

# save RAM
del train_ds, history


stride = config["stride"]
if False:  #'lustre' in config['datagen'].lower()):
    logging.info("Making predictions on validation data...")
    # preallocate
    # Making a 1D array with the number of pixels/predictions we'll have.
    # First find how many we'll have. Only works if using lustre samplelists.
    #  npix = 0
    #  for d in val_list:
    #    npixY = (d['endY'] - d['startY'])//stride
    #    npixX = (d['endX'] - d['startX'])//stride
    #    npix += (npixY * npixX)
    #  preds = np.zeros(npix,dtype=np.float16)
    #  val_target = np.zeros(npix,dtype=np.float16)
    #  curr_start_idx = 0

    ny, nx = config["ps_2km"]
    remainder = n_vsamples % val_bs
    preds = np.empty((n_vsamples, ny // stride, nx // stride), dtype=np.float32)
    val_target = np.empty((n_vsamples, ny // stride, nx // stride), dtype=np.float32)

    t0 = time.time()
    cter = 0
    for (
        inputs,
        labels,
    ) in val_ds:  # shuffle should be False, especially if spatial_verification == True
        print(cter)
        try:
            tmp_preds = conv_model.predict(inputs)
            preds[cter * val_bs : (cter + 1) * val_bs] = tmp_preds[
                :, ::stride, ::stride
            ]
            val_target[cter * val_bs : (cter + 1) * val_bs] = labels[
                :, ::stride, ::stride, 0
            ]
        except ValueError:  # the remainder
            tmp_preds = conv_model.predict(inputs)
            preds[cter * val_bs : cter * val_bs + remainder] = tmp_preds[
                :, ::stride, ::stride
            ]
            val_target[cter * val_bs : cter * val_bs + remainder] = labels[
                :, ::stride, ::stride, 0
            ]
        cter += 1

    preds = preds.flatten()
    val_target = val_target.flatten()
    print("validation computation: ", time.time() - t0)
    print(len(preds), len(val_target))

    assert len(preds) == len(val_target)
    print(preds.max(), preds.min())
    print(val_target.max(), val_target.min())

    scores_dict = vv.verification(preds, val_target, outdir)
    pickle.dump(scores_dict, open(os.path.join(outdir, "scores_dict.pkl"), "wb"))
#####################################################################################################
#####################################################################################################


add_scores_to_file(outdir)
