import torch
import netCDF4
import numpy as np
import os, sys
from lightningcast import solar_angles
from datetime import datetime

# -------------------------------------------------------------------------------------------------#
class DatasetNETCDF(torch.utils.data.Dataset):
    "Characterizes a dataset for PyTorch"

    def __init__(self, filelist, config):
        "Initialization"
        self.filelist = filelist
        self.inputs = config["inputs"]
        self.scaling_vals = config["scaling_vals"]
        # self.nframes = config['nframes']
        self.target = config["target"]
        self.binarize = config["binarize"]
        self.scalar_vars = config["scalar_vars"]
        self.scalar_tuple = config["scalar_tuple"]
        self.input_tuples = config["input_tuples"]

    def __len__(self):
        "Denotes the total number of samples"
        return len(self.filelist)

    def __getitem__(self, index):
        "Generate one sample of data"
        file = self.filelist[index]
        inputs = []
        nc = netCDF4.Dataset(file, "r")

        for inp in self.inputs:  # a list of lists (non-scalar vars)
            X = None
            for chan in inp:
                tmp = np.expand_dims(
                    (nc.variables[chan][-1, ...] - self.scaling_vals[chan]["mean"])
                    / self.scaling_vals[chan]["std"],
                    axis=0,
                )  # make channels axis
                tmp = tmp.filled(
                    self.scaling_vals[chan]["mean"]
                )  # also converts to numpy.ndarray
                if X is None:
                    X = tmp + 0.0
                else:
                    X = np.concatenate((X, tmp), axis=0)  # need channels first
            inputs.append(X)

        if len(self.scalar_vars):
            sy = self.scalar_tuple[-2]
            sx = self.scalar_tuple[-1]  # last input_tuple should be scalars
            X = None
            for svar in self.scalar_vars:
                if svar == "month":
                    # e.g. goes16_conus_20180707-001722.netcdf
                    dateinfo = os.path.basename(file).split("conus_")[-1][
                        0:8
                    ]  # yyyymmdd
                    sv = datetime.strptime(dateinfo, "%Y%m%d").month
                    tmp = np.full((1, sy, sx), sv)  # make same value everywhere
                    tmp_scaled = (
                        tmp - self.scaling_vals[svar]["mean"]
                    ) / self.scaling_vals[svar]["std"]
                if X is None:
                    X = tmp_scaled + 0.0
                else:
                    X = np.concatenate((X, tmp_scaled), axis=0)  # need channels first
            inputs.append(X)

        # targets
        y = np.array(np.expand_dims(nc.variables[self.target][:], axis=0))
        y[y < self.binarize] = 0
        y[y >= self.binarize] = 1

        nc.close()

        return inputs, y


# -------------------------------------------------------------------------------------------------#
class DatasetLUSTRE(torch.utils.data.Dataset):
    "Characterizes a dataset for PyTorch"

    def __init__(self, samplelist, config, lats=None, lons=None, satzen=None):
        "Initialization"
        self.samplelist = samplelist  # this is a list of dicts with info on each sample
        self.inputs = config["inputs"]
        self.scaling_vals = config["scaling_vals"]
        self.target = config["target"]
        self.binarize = config["binarize"]
        self.scalar_vars = config["scalar_vars"]
        self.scalar_tuple = config["scalar_tuple"]
        self.input_tuples = config["input_tuples"]
        self.srrf = config["scalar_reduce_res_fac"]  # usu. 4
        self.nclass = config["nclass"]
        self.lats = lats
        self.lons = lons
        self.satzen = satzen

    def __len__(self):
        "Denotes the total number of samples"
        return len(self.samplelist)

    def __getitem__(self, index):
        """
        Generate one sample of data.
        Keys in each sample:
        - 'abidt': <datetime.datetime> for ABI scan
        - 'startX','endX','startY','endY': <int> The coordinate
          endpoints in image space, for *2-km* ABI data.
          With RadC files, this is from the fixed-grid format.
          We are breaking up CONUS scenes, so these will vary,
          with about 6 sets of corner points per ABI dt.
        - 'target_file': <str> Full file name for the accumlated/aggregated
          truth metric. E.g., in /apollo/grain/...../FED_accum_60min/%Y%m%d-%H%M00.netcdf
        - 'CH02'...'CH16': <str> The full filenames for each channel. I think it
          is better to just put the file names here, "precomputed" once
          rather than globbing each epoch.
        """

        sample = self.samplelist[index]
        inputs = []
        startY = sample["startY"]
        endY = sample["endY"]
        startX = sample["startX"]
        endX = sample["endX"]

        for inp in self.inputs:  # a list of lists (non-scalar vars)
            X = None
            for chan in inp:
                abifile = sample[chan]

                tmp = self._make_abi_data(
                    abifile,
                    ch=chan[2:],
                    startX=startX,
                    endX=endX,
                    startY=startY,
                    endY=endY,
                )

                # tmp = np.ma.fix_invalid(tmp) #masks nan/inf; currently unnecessary, as data are properly masked
                tmp_scaled = np.expand_dims(
                    (tmp - self.scaling_vals[chan]["mean"])
                    / self.scaling_vals[chan]["std"],
                    axis=0,
                )  # make channels axis
                tmp_scaled = tmp_scaled.filled(
                    self.scaling_vals[chan]["mean"]
                )  # also converts to numpy.ndarray
                if X is None:
                    X = tmp_scaled + 0.0
                else:
                    X = np.concatenate((X, tmp_scaled), axis=0)  # need channels first
            inputs.append(X)

        # make scalars -- note that in a U-Net, these aren't strictly 'scalars'
        if len(self.scalar_vars):
            sy = self.scalar_tuple[-2]
            sx = self.scalar_tuple[-1]  # last input_tuple should be scalars
            X = None
            for svar in self.scalar_vars:
                if svar == "month":
                    tmp = np.full(
                        (sy, sx), sample["abidt"].month
                    )  # make same value everywhere
                elif svar == "week_of_year":
                    tmp = np.full(
                        (sy, sx), int(sample["abidt"].strftime("%j")) // 7
                    )  # make same value everywhere
                elif svar == "fortnight_of_year":
                    tmp = np.full(
                        (sy, sx), int(sample["abidt"].strftime("%j")) // 14
                    )  # make same value everywhere
                elif svar == "cosday":  # should be used with sinday
                    doy = int(sample["abidt"].strftime("%j"))
                    tmp = np.full(
                        (sy, sx), np.cos((2 * np.pi / 365) * doy)
                    )  # make same value everywhere # y = cos((2*pi/365) * x), for cylical value
                    # i.e., Dec31 is like Jan1
                elif svar == "sinday":  # should be used with cosday
                    doy = int(sample["abidt"].strftime("%j"))
                    tmp = np.full(
                        (sy, sx), np.sin((2 * np.pi / 365) * doy)
                    )  # make same value everywhere # y = sin((2*pi/365) * x), for cylical value
                    # i.e., Dec31 is like Jan1
                elif svar == "solzen":
                    # assuming we want to reduce resolution. start/end X/Y are for 2-km CONUS FGF
                    lat = self.lats[
                        startY // self.srrf : endY // self.srrf,
                        startX // self.srrf : endX // self.srrf,
                    ]
                    lon = self.lons[
                        startY // self.srrf : endY // self.srrf,
                        startX // self.srrf : endX // self.srrf,
                    ]
                    tmp, _ = solar_angles.solar_angles(
                        dt=[sample["abidt"]], lat=lat, lon=lon
                    )
                elif svar == "lat":
                    # assuming we want to reduce resolution.
                    tmp = self.lats[
                        startY // self.srrf : endY // self.srrf,
                        startX // self.srrf : endX // self.srrf,
                    ]
                elif svar == "lon":
                    # assuming we want to reduce resolution.
                    tmp = self.lons[
                        startY // self.srrf : endY // self.srrf,
                        startX // self.srrf : endX // self.srrf,
                    ]
                elif svar == "satzen":
                    # assuming we want to reduce resolution.
                    tmp = self.satzen[
                        startY // self.srrf : endY // self.srrf,
                        startX // self.srrf : endX // self.srrf,
                    ]
                tmp = np.ma.fix_invalid(tmp)  # masks nan/inf
                tmp_scaled = np.expand_dims(
                    (tmp - self.scaling_vals[svar]["mean"])
                    / self.scaling_vals[svar]["std"],
                    axis=0,
                )  # scale and expand
                try:
                    tmp_scaled = tmp_scaled.filled(
                        self.scaling_vals[svar]["mean"]
                    )  # fill bad pixels with mean...also converts to numpy.ndarray
                except AttributeError:
                    pass  # it's already an ndarray, not a MaskedArray
                if X is None:
                    X = tmp_scaled + 0.0
                else:
                    X = np.concatenate((X, tmp_scaled), axis=0)  # need channels first
            inputs.append(X)

        # target/truth
        target_file = sample["target_file"]
        nc = netCDF4.Dataset(target_file, "r")
        # check if we're using 8-km data
        if nc.spatial_resolution == "reduced from 2-km to 8-km":
            y = np.expand_dims(
                nc.variables[self.target][
                    startY // self.srrf : endY // self.srrf,
                    startX // self.srrf : endX // self.srrf,
                ],
                axis=0,
            )
            increase_res = True
        else:  # else assume 2-km target data
            if self.nclass == 1:  # binary cross-entropy
                y = np.expand_dims(
                    nc.variables[self.target][startY:endY, startX:endX], axis=0
                )
            else:  # for cross-entropy, I guess the channel dim is not needed
                y = nc.variables[self.target][startY:endY, startX:endX]
            increase_res = False
        y = y.filled(
            0
        )  # masked values become 0 and y becomes np.ndarray instead of np.ma.core.MaskedArray (required by pytorch)
        y[y < self.binarize] = 0
        y[y >= self.binarize] = 1
        if increase_res:
            y = y.repeat(4, axis=1).repeat(
                4, axis=2
            )  # artificially restore to 2-km res. 0th is channels dim.

        nc.close()
        return inputs, y

    def _make_abi_data(
        self, abifile, ch=None, startY=0, endY=1500, startX=0, endX=2500
    ):
        "Return the possibly cropped reflectance or BT"
        nc = netCDF4.Dataset(abifile, "r")
        if ch is None:  # get chan from file name
            basefile = os.path.basename(abifile)
            ch = basefile.split("_")[1].split("-")[-1].split("C")[-1]

        if ch == "02":
            # convert vis radiance to reflectance
            result = (
                nc.variables["Rad"][startY * 4 : endY * 4, startX * 4 : endX * 4]
                * nc.variables["kappa0"][:]
            )
        elif ch in ["01", "03", "05"]:
            # convert vis radiance to reflectance
            result = (
                nc.variables["Rad"][startY * 2 : endY * 2, startX * 2 : endX * 2]
                * nc.variables["kappa0"][:]
            )
        else:  # IR channels
            # convert IR radiance to BT
            radiance = nc.variables["Rad"][startY:endY, startX:endX]
            result = (
                nc.variables["planck_fk2"][:]
                / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
                - nc.variables["planck_bc1"][:]
            ) / nc.variables["planck_bc2"][:]
        nc.close()
        return result


# -------------------------------------------------------------------------------------------------#
class DatasetZARR_UNET(torch.utils.data.Dataset):
    "Characterizes a dataset for PyTorch"

    def __init__(self, inds, config, zarr=None, lats=None, lons=None, satzen=None):
        "Initialization"
        self.inds = inds
        self.inputs = config["inputs"]
        self.scaling_vals = config["scaling_vals"]
        self.target = config["target"]
        self.binarize = config["binarize"]
        self.scalar_vars = config["scalar_vars"]
        self.scalar_tuple = config["scalar_tuple"]
        self.input_tuples = config["input_tuples"]
        self.zarr = zarr
        self.srrf = config["scalar_reduce_res_fac"]  # usu. 4
        self.lats = lats
        self.lons = lons
        self.satzen = satzen

    def __len__(self):
        "Denotes the total number of samples"
        return len(self.inds)

    def __getitem__(self, index):

        inputs = []

        for inp in self.inputs:  # a list of lists (non-scalar vars)
            X = None
            for chan in inp:
                tmp = self.zarr[chan].get_orthogonal_selection(
                    ([index], [-1], slice(None), slice(None), [-1])
                )
                # here, we're keeping the time axis to serve as a channel axis. Remove the 'channel' axis from the zarr data.
                tmp = np.squeeze(tmp, axis=(0, -1))  # Don't need sample axis either
                tmp_scaled = (
                    tmp - self.scaling_vals[chan]["mean"]
                ) / self.scaling_vals[chan]["std"]
                # tmp_scaled = tmp_scaled.filled(self.scaling_vals[chan]['mean']) #also converts to numpy.ndarray
                if X is None:
                    X = tmp_scaled + 0.0
                else:
                    X = np.concatenate((X, tmp_scaled), axis=0)  # need channels first
            inputs.append(X)

        # make target data
        yslice = slice(None)  # yslice = list(range(400,800))
        xslice = slice(None)  # xslice = list(range(0,400))
        y = self.zarr[self.target].get_orthogonal_selection(
            ([index], yslice, xslice, [-1])
        )
        y[y < self.binarize] = 0
        y[y >= self.binarize] = 1
        y = np.squeeze(y, axis=-1)  # remove the superfluous last dim

        # print(inputs[0].shape,inputs[1].shape,inputs[2].shape)
        # print(y.shape)
        return inputs, y
